import numpy as np
import hypothesis as hyp
#from hypothesis import given
import hypothesis.strategies as st
import pytest
#import sys
#sys.path.append("c:/skola/utokyo_lab/ancsim/ancsim")

from aspsim.simulator import SimulatorSetup
import aspsim.array as ar
import aspsim.processor as bse
import aspsim.diagnostics.core as diacore
import aspsim.diagnostics.diagnostics as dia
import aspsim.fileutilities as fu
import aspsim.signal.sources as sources


@pytest.fixture(scope="session")
def fig_folder(tmp_path_factory):
    return tmp_path_factory.mktemp("figs")

def simple_setup(fig_folder):
    setup = SimulatorSetup(fig_folder)
    setup.sim_info.tot_samples = 20
    setup.sim_info.sim_buffer = 20
    setup.sim_info.export_frequency = 20
    setup.sim_info.sim_chunk_size = 20

    setup.add_free_source("src", np.array([[1,0,0]]), sources.WhiteNoiseSource(1,1))
    setup.add_controllable_source("loudspeaker", np.array([[1,0,0]]))
    setup.add_mics("mic", np.array([[0,0,0]]))

    setup.arrays.path_type["loudspeaker"]["mic"] = "none"
    setup.arrays.path_type["src"]["mic"] = "direct"
    return setup

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=10), 
           export_freq = st.integers(min_value=1, max_value=10))
def test_processor_sees_same_mic_samples_as_is_logged_in_record_signal(fig_folder, bs, export_freq):
    setup = simple_setup(fig_folder)
    setup.sim_info.export_frequency = export_freq

    sim = setup.create_simulator()
    proc = bse.DebugProcessor(sim.sim_info, sim.arrays, bs)
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, export_func="npz"))
    sim.add_processor(proc)
    sim.run_simulation()

    final_export_idx = sim.sim_info.export_frequency * (sim.sim_info.tot_samples // sim.sim_info.export_frequency)
    signal_log = np.load(sim.folder_path.joinpath(f"mic_{final_export_idx}.npz"))["mic"]
    signal_true = sim.processors[0].mic
    assert np.allclose(signal_log, signal_true[:,:final_export_idx])



@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=10))
def test_record_signal_is_same_with_or_without_a_processor(fig_folder, bs):
    setup = simple_setup(fig_folder)
    final_export_idx = setup.sim_info.export_frequency * (setup.sim_info.tot_samples // setup.sim_info.export_frequency)

    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, export_func="npz"))
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()

    sig_with_proc = np.load(sim.folder_path.joinpath(f"mic_{final_export_idx}.npz"))["mic"]

    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, export_func="npz"))
    sim.run_simulation()

    sig_without_proc = np.load(sim.folder_path.joinpath(f"mic_{final_export_idx}.npz"))["mic"]

    assert np.allclose(sig_with_proc, sig_without_proc)


@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=10),
           export_freq = st.integers(min_value=2, max_value=10),
           tot_samples = st.integers(min_value=10, max_value=30),
           )
def test_signal_diagnostics_correct_files_saved(fig_folder, bs, export_freq, tot_samples):
    sim_setup = simple_setup(fig_folder)
    sim_setup.sim_info.tot_samples = tot_samples
    sim_setup.sim_info.export_frequency = export_freq
    sim_setup.sim_info.sim_chunk_size = 5
    sim_setup.sim_info.sim_buffer = 20

    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, export_func="npz"))
    sim.run_simulation()

    all_saved_files = list(sim.folder_path.iterdir())
    num_files_to_save = sim.sim_info.tot_samples // sim_setup.sim_info.export_frequency
    expected_files = [sim.folder_path.joinpath(f"mic_{i*sim_setup.sim_info.export_frequency}.npz") for i in range(1, num_files_to_save+1)]

    # Check all expected files exist
    for f in expected_files:
        assert f.exists()

    # Check there are no other npz files except the expected files
    for f in all_saved_files:
        if f.suffix == ".npz":
            assert f in expected_files


@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5),
            buf_size = st.integers(min_value=10, max_value=30),
            num_proc = st.integers(min_value=1, max_value=3))
def test_all_samples_saved_for_signal_diagnostics(fig_folder, bs, buf_size, num_proc):
    sim_setup = simple_setup(fig_folder)
    sim_setup.sim_info.sim_buffer = buf_size
    sim = sim_setup.create_simulator()
    for _ in range(num_proc):
        sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                    diagnostics={"mic":dia.RecordSignal("mic", sim.sim_info, bs, export_func="npz")}))

    sim.run_simulation()
    
    at_least_one_file_saved = False
    for f in sim.folder_path.iterdir():
        if f.stem.startswith("mic"):
            at_least_one_file_saved = True
            saved_data = np.load(f)
            for proc_name, data in saved_data.items():
                assert np.allclose(data, np.arange(sim.sim_info.sim_buffer, 
                            sim.sim_info.sim_buffer+sim.sim_info.tot_samples))
    assert at_least_one_file_saved



@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5),
            buf_size = st.integers(min_value=10, max_value=30))
def test_correct_intermediate_samples_saved_for_signal_diagnostics(fig_folder, bs, buf_size):
    sim_setup = simple_setup(fig_folder)
    sim_setup.sim_info.sim_buffer = buf_size
    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                   diagnostics={"mic":dia.RecordSignal("mic", sim.sim_info, bs, export_func="npz", keep_only_last_export=False)}))

    sim.run_simulation()
    
    for f in sim.folder_path.iterdir():
        if f.stem.startswith("mic"):
            idx = fu.find_index_in_name(f.stem)
            saved_data = np.load(f)
            for proc_name, data in saved_data.items():
                assert np.allclose(data, np.arange(sim.sim_info.sim_buffer, 
                            sim.sim_info.sim_buffer+idx))
    

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_export_file_naming_interval_diagnostics(fig_folder, bs):
    sim_setup = simple_setup(fig_folder)
    sim = sim_setup.create_simulator()

    save_intervals = ((32,46), (68,69), (71, 99))
    diag_name = "mic"
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                   diagnostics={"mic":dia.RecordSignal(diag_name, sim.sim_info, bs, 
                   export_at = [iv[1] for iv in save_intervals],
                    save_at=diacore.IntervalCounter(save_intervals), 
                    export_func="npz")}))
    sim.run_simulation()

    for iv in save_intervals:
        assert sim.folder_path.joinpath(f"{diag_name}_{iv[1]}.npz").exists()



@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_correct_samples_saved_for_interval_diagnostics(fig_folder, bs):
    sim_setup = simple_setup(fig_folder)
    sim = sim_setup.create_simulator()

    save_intervals = ((32,46), (68,69), (71, 99))
    diag_name = "mic"
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                   diagnostics={"mic":dia.RecordSignal(
                       diag_name, sim.sim_info, bs, 
                   export_at = [iv[1] for iv in save_intervals],
                    save_at = diacore.IntervalCounter(save_intervals), 
                    export_func="npz")
                    }))

    sim.run_simulation()

    expected = np.zeros(0)
    for iv in save_intervals:
        saved_data = np.load(sim.folder_path.joinpath(f"{diag_name}_{iv[1]}.npz"))
        for proc_name, data in saved_data.items():
            #expected[iv[0]:iv[1]] = np.arange(iv[0]+sim.sim_info.sim_buffer, 
            #                                    iv[1]+sim.sim_info.sim_buffer)
            expected = np.concatenate((expected, np.arange(iv[0]+sim.sim_info.sim_buffer, 
                                                        iv[1]+sim.sim_info.sim_buffer)))

            assert np.allclose(
                data,
                expected,
                equal_nan=True
            )
            



@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5),
            buf_size = st.integers(min_value=10, max_value=30),
            num_proc = st.integers(min_value=1, max_value=3))
def test_all_samples_saved_state_diagnostics(fig_folder, bs, buf_size, num_proc):
    sim_setup = simple_setup(fig_folder)
    #bs = 1
    #sim_setup.sim_info.tot_samples = 13
    #sim_setup.sim_info.sim_chunk_size = 5
    sim_setup.sim_info.sim_buffer = buf_size

    sim = sim_setup.create_simulator()

    #save_at = diacore.IntervalCounter(np.arange(1,sim.sim_info.tot_samples+1))
    for _ in range(num_proc):
        sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                    diagnostics={"state":dia.RecordState("processed_samples", 1, sim.sim_info, bs, export_func="npz")}))

    sim.run_simulation()
    
    one_file_saved = False
    for f in sim.folder_path.iterdir():
        if f.stem.startswith("state"):
            one_file_saved = True
            idx = fu.find_index_in_name(f.stem)
            saved_data = np.load(f)
            for proc_name, data in saved_data.items():
                assert np.allclose(data, np.arange(bs, idx+1, bs))
    assert one_file_saved


@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_correct_samples_saved_for_instant_diagnostics(fig_folder, bs):
    sim_setup = simple_setup(fig_folder)
    sim = sim_setup.create_simulator()

    #save_at = np.arange(bs, sim.sim_info.tot_samples, bs)#(bs,)
    #save_at = [bs*i for i in range(1, sim.sim_info.tot_samples//bs)]
    save_at = (bs, 2*bs, 5*bs)
    #save_intervals = ((1,2), (3,4), (5,6))
    diag_name = "filt"
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                   diagnostics={diag_name:dia.RecordFilter(
                       "filt.ir", sim.sim_info, bs, save_at = save_at, export_func="npz")}))
                #        "mic":dia.RecordSignal(
                #        "mic", sim.sim_info, bs, 
                #    export_at = [iv[1] for iv in save_intervals],
                #     save_at = diacore.IntervalCounter(save_intervals), 
                #     export_func="npz")
                #     }))

    sim.run_simulation()

    for idx in save_at:
        saved_data = np.load(sim.folder_path.joinpath(f"{diag_name}_{idx}.npz"))
        for proc_name, data in saved_data.items():
            assert np.allclose(data, np.zeros_like(data)+idx)

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_correct_samples_saved_for_instant_diagnostics_savefreq(fig_folder, bs):
    sim_setup = simple_setup(fig_folder)
    sim = sim_setup.create_simulator()

    save_at = bs
    diag_name = "filt"
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
                   diagnostics={diag_name:dia.RecordFilter(
                       "filt.ir", sim.sim_info, bs, save_at = save_at, export_func="npz")}))

    sim.run_simulation()

    for idx in range(save_at, sim.sim_info.tot_samples+1, save_at):
        saved_data = np.load(sim.folder_path.joinpath(f"{diag_name}_{idx}.npz"))
        for proc_name, data in saved_data.items():
            assert np.allclose(data, np.zeros_like(data)+idx)



@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_two_processors_with_different_diagnostics(fig_folder, bs):
    sim_setup = simple_setup(fig_folder)
    sim = sim_setup.create_simulator()

    proc1 = bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
            diagnostics = {"common" : dia.RecordSignal("mic", sim.sim_info, bs, export_func = "npz", keep_only_last_export=False),
                            "individual1" : dia.RecordSignal("mic", sim.sim_info, bs, export_func = "npz", keep_only_last_export=False),
            }
            )
    proc2 = bse.DebugProcessor(sim.sim_info, sim.arrays, bs, 
            diagnostics = {"common" : dia.RecordSignal("mic", sim.sim_info, bs, export_func = "npz", keep_only_last_export=False),
                            "individual2" : dia.RecordSignal("mic", sim.sim_info, bs, export_func = "npz", keep_only_last_export=False)
                }
            )

    sim.add_processor(proc1)
    sim.add_processor(proc2)
    sim.run_simulation()

    for f in sim.folder_path.iterdir():
        if f.stem.startswith("mic"):
            idx = fu.find_index_in_name(f.stem)
            saved_data = np.load(f)
            for proc_name, data in saved_data.items():
                assert np.allclose(data[:idx+1], np.arange(sim.sim_info.sim_buffer, 
                            sim.sim_info.sim_buffer+idx+1))




import numpy as np
import copy

import aspsim.fileutilities as futil
import aspsim.configutil as configutil
import aspsim.array as ar
import aspsim.saveloadsession as sess
import aspsim.diagnostics.core as diacore

import aspcore.filter as fc


class SimulatorSetup:
    def __init__(
        self,
        base_fig_path=None,
        session_folder=None,
        config_path = None,
        rng = None, 
        ):
        """Use this class to set up a simulation. 

        Unless you really know what you are doing, you should use this 
        and not the Simulator class directly. After all parameters and arrays are
        set, call create_simulator() to obtain a Simulator object.

        Parameters
        ----------
        base_fig_path : str or Path from pathlib
            If this is supplied, the simulator will create a new
            subfolder in that directory and fill it with plots and metadata
        session_folder : str or Path from pathlib
            If the option auto_save_load in the config is True, the simulator will 
            look in the session_folder for sessions to load, and will save the current
            session there. 
        config_path : str or Path from pathlib
            Supply if you want to load the config parameters from a (yaml) file. 
            Otherwise the default will be loaded, which can be changed inside your
            Python code. 
        rng : numpy Generator object
            recommended to obtain your rng object from np.random.default_rng(seed)
        """
        if config_path is None:
            self.sim_info = configutil.load_default_config()
        else:
            self.sim_info = configutil.load_from_file(config_path)

        self.base_fig_path = base_fig_path
        self.session_folder = session_folder

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

        self.arrays = ar.ArrayCollection()
        #self.diag = diacore.DiagnosticHandler(self.sim_info) 

    def add_arrays(self, array_collection):
        """Adds all arrays and path types from an array collection
        
        Parameters
        ----------
        array_collection : array.ArrayCollection
        """
        for ar in array_collection:
            self.arrays.add_array(ar)
        self.arrays.set_path_types(array_collection.path_type)

            #if ar.is_source:
            #    self.arrays.paths[ar.name] = {}

    def add_array(self, array):
        """Add an array to the simulation.
        
        Parameters
        ----------
        array : array object
            The array to add, subclass of Array. 
            See the array module for examples.
        """
        self.arrays.add_array(array)

    def add_free_source(self, name, pos, source):
        """Adds a free source array at the given position.
        
        Parameters
        ----------
        name : str
            Name of the array
        pos : ndarray of shape (num_pos, spatial_dim)
            Positions of the sources in the array
        source : source object
            The sound source to use for the array. See the sources module for examples.
            Can in principle be any object with a get_samples(num_samples) method 
            that returns an ndarray of shape (num_pos, num_samples)
        """
        arr = ar.FreeSourceArray(name, pos, source)
        self.add_array(arr)

    def add_controllable_source(self, name, pos):
        """Adds a controllable source array at the given position.
        
        Parameters
        ----------
        name : str
            Name of the array
        pos : ndarray of shape (num_pos, spatial_dim)
            Positions of the sources in the array
        """
        arr = ar.ControllableSourceArray(name, pos)
        self.add_array(arr)
    
    def add_mics(self, name, pos, **kwargs):
        """Adds a microphone array at the given position.
        
        Parameters
        ----------
        name : str
            Name of the array
        pos : ndarray of shape (num_pos, spatial_dim)
            Positions of the mics in the array
        """
        arr = ar.MicArray(name, pos, **kwargs)
        self.add_array(arr)

    def set_path(self, src_name, mic_name, path):
        """Sets the path between a source and a microphone array.
        
        Parameters
        ----------
        src_name : str
            Name of the source array
        mic_name : str
            Name of the microphone array
        path : ndarray of shape (num_src, num_mic, path_len)
            The path between the source and the microphone array.
        """
        self.arrays.set_prop_path(path, src_name, mic_name)

    def set_source(self, name, source):
        """Adds a source to a FreeSourceArray.
        
        Will raise an error if the array is not a FreeSourceArray or 
        is not added to the simulation yet.

        Parameters
        ----------
        name : str
            Name of the array
        source : source object
            see sources module for examples
        """
        self.arrays[name].set_source(source)

    def load_from_path(self, session_path):
        """Loads simulator setup from a previous session in session_path.
        
        Parameters
        ----------
        session_path : str or Path from pathlib
        """
        self.folder_path = self._create_fig_folder(self.base_fig_path)
        self.sim_info, self.arrays = sess.load_from_path(session_path, self.folder_path)

    def create_simulator(self):
        """Creates and returns a simulator object from the current setup.

        If fig_path is set, the simulator will create a new subfolder and save
        metadata about the simulation parameters.

        If possible, the simulator will load a previous session from the session_folder
        
        Returns
        -------
        simulator : Simulator object

        Notes
        -----
        Can be used multiple times with changes to the config in between to 
        create similar simulations with different parameters.
        """
        assert not self.arrays.empty()
        finished_arrays = copy.deepcopy(self.arrays)
        finished_arrays.set_default_path_type(self.sim_info.reverb)
        
        folder_path = self._create_fig_folder(self.base_fig_path)
        print(f"Figure folder: {folder_path}")

        if self.sim_info.auto_save_load and self.session_folder is not None:
            try:
                finished_arrays = sess.load_session(self.session_folder, folder_path, self.sim_info, finished_arrays)
            except sess.MatchingSessionNotFoundError:
                print("No matching session found")
                ir_metadata = finished_arrays.setup_ir(self.sim_info)
                sess.save_session(self.session_folder, self.sim_info, finished_arrays, sim_metadata=ir_metadata)   
        else:
            finished_arrays.setup_ir(self.sim_info)

        # LOGGING AND DIAGNOSTICS
        self.sim_info.save_to_file(folder_path)
        finished_arrays.plot(self.sim_info, folder_path, self.sim_info.plot_output)
        finished_arrays.save_metadata(folder_path)
        return Simulator(self.sim_info, finished_arrays, folder_path, rng=self.rng)

    def _create_fig_folder(self, folder_for_plots, gen_subfolder=True, safe_naming=False):
        if self.sim_info.plot_output == "none" or folder_for_plots is None:
            return None

        if gen_subfolder:
            folder_name = futil.get_unique_folder_name("figs_", folder_for_plots, safe_naming)
            folder_name.mkdir()
        else:
            folder_name = folder_for_plots
            if not folder_name.exists():
                folder_name.mkdir()
        return folder_name


class Simulator:
    def __init__(
        self,
        sim_info,
        arrays,
        folder_path,
        rng = None, 
    ):
        
        self.sim_info = sim_info
        self.arrays = arrays
        self.folder_path = folder_path
        self.diag = diacore.Logger(self.sim_info)
        
        self.processors = []

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

    def add_processor(self, processor):
        """Add one or several processors to the simulation.

        Parameters
        ----------
        processor : processor object or list of processor objects
            See the processors module for examples.
        """
        try:
            self.processors.extend(processor)
        except TypeError:
            self.processors.append(processor)

    def _prepare_simulation(self):
        if len(self.processors) > 1:
            raise NotImplementedError("Multiple processors are not yet supported in the simulator.")
        
        set_unique_processor_names(self.processors)
        sess.write_processor_metadata(self.processors, self.folder_path)

        
        #self.plot_exporter = diacore.DiagnosticExporter(
        #    self.sim_info, self.diag
        #)

        self.sig = Signals(self.sim_info, self.arrays)
        local_signals = [sig_name for sig_name in self.processors[0].sig.keys() if sig_name not in self.sig]
        for sig_name in local_signals:
            self.sig.create_signal(sig_name, self.processors[0].sig[sig_name].shape[:-1])


        self.propagator = Propagator(self.sim_info, self.arrays, self.sig)

        self.diag.prepare()
        self.propagator.prepare()
        for proc in self.processors:
            proc.sig = self.sig
            proc.prepare()
            

    def run_simulation(self):
        self._prepare_simulation()

        # the else value (which happens when there is no processor) can be changed to just about anything. Could be set to 1, or 
        # set to a high value that can be used to speed up. 
        max_block_size = np.max([proc.block_size for proc in self.processors]) if self.processors else self.sim_info.sim_buffer // 4  

        print("SIM START")
        self.n_tot = 0
        while self.n_tot < self.sim_info.tot_samples+max_block_size:
            self.arrays.update(self.n_tot)

            for proc in self.processors:
                if self.n_tot % proc.block_size == proc.block_size-1:
                    proc.process(proc.block_size)
            self.propagator.propagate(1)
            self.diag_moved_from_propagate(max_block_size)

            self.diag.dispatch(self.folder_path)

            # Write progress
            if self.n_tot % 1000 == 0:
                print(f"Timestep: {self.n_tot}")#, end="\r")

            self.sig.idx += 1
            self.n_tot += 1
        self.diag.dispatch(self.folder_path)

        print(self.n_tot)

    def diag_moved_from_propagate(self, max_block_size):
        #Temporary, especially the function name
        last_block = self.last_block_on_buffer(max_block_size)

        if len(self.processors) > 1:
            raise NotImplementedError("Multiple processors are not yet supported in the simulator.")
        self.diag.save_data(self.processors[0], self.sig, self.sig.idx, self.n_tot, last_block)
        if last_block:
            self.sig._reset_signals()

    def last_block_on_buffer(self, max_block_size):
        return self.sig.idx+max_block_size >= self.sim_info.sim_chunk_size+self.sim_info.sim_buffer

def set_unique_processor_names(processors):
    names = []
    for proc in processors:
        new_name = proc.name
        i = 1
        while new_name in names:
            new_name = f"{proc.name} {i}"
            i += 1
        names.append(new_name)
        proc.name = new_name





class Signals():
    """
    A simple wrapper around a dictionary to hold all signals
    """
    def __init__(self, sim_info, arrays):
        self.sim_info = sim_info
        self.signals = {}
        self.idx = 0
        self.idx_tot = 0

        for array in arrays:
            self.create_signal(array.name, array.num)
        if self.sim_info.save_source_contributions:
            for src, mic in arrays.mic_src_combos():
                self.create_signal(src.name+"~"+mic.name, mic.num)
      
    def __getitem__(self, key):
        return self.signals[key]

    def __contains__(self, key):
        return key in self.signals

    def items(self):
        return self.signals.items()
    
    def values(self):
        return self.signals.values()

    def keys(self):
        return self.signals.keys()

    def create_signal(self, name, dim):
        """
        Inserts a new signal of shape (*dim, simbuffer+simchunksize)

        Parameters
        ----------
        name : str 
            name of the signal, in the same way a dictionary entry has a name
            signal is accessed as signals[name]
        dim : int, list or tuple
            dimensionality of the signal. signal will have the shape (*dim, simbuffer+simchunksize)
            where the last axis represents time
        """
        if isinstance(dim, int):
            dim = (dim,)
        assert name not in self.signals
        self.signals[name] = np.zeros(dim + (self.sim_info.sim_buffer + self.sim_info.sim_chunk_size,))

    def _reset_signals(self):
        for name, sig in self.signals.items():
            self.signals[name] = np.concatenate(
                (
                    sig[..., -self.sim_info.sim_buffer :],
                    np.zeros(sig.shape[:-1] + (self.sim_info.sim_chunk_size,)),
                ),
                axis=-1,
            )
        self.idx -= self.sim_info.sim_chunk_size


class Propagator():
    def __init__(self, sim_info, arrays, sig):
        self.sim_info = sim_info
        self.arrays = arrays
        self.sig = sig

        self.path_filters = {}
        for src, mic, path in arrays.iter_paths():
            if src.name not in self.path_filters:
                self.path_filters[src.name] = {}
            self.path_filters[src.name][mic.name] = fc.create_filter(ir=path, sum_over_input=True, dynamic=(src.dynamic or mic.dynamic))

    def prepare(self):
        """Prepares the initial state of the signals.

        If start_sources_before_0 is True, the source signals will have started sim_buffer samples before time 0. This can be
        useful if you want to have a stationary signal at time 0.

        If start_sources_before_0 is False, the source signals will start at time 0. Therefore this function only generates
        a single sample of the source signals. 
        
        Currently does not take movement into account. It just propagates from the stationary RIRs associated with the initial position
        """
        num_samples = self.sim_info.sim_buffer if self.sim_info.start_sources_before_0 else 1
        end_sample = self.sim_info.sim_buffer

        for src in self.arrays.free_sources():
            self.sig[src.name][..., end_sample - num_samples:end_sample] = src.get_samples(num_samples)

        for src, mic in self.arrays.mic_src_combos():
            propagated_signal = self.path_filters[src.name][mic.name].process(
                    self.sig[src.name][:, end_sample - num_samples:end_sample])
            self.sig[mic.name][:,:self.sim_info.sim_buffer] += propagated_signal
            if self.sim_info.save_source_contributions:
                self.sig[f"{src.name}~{mic.name}"][:,end_sample - num_samples:end_sample] = propagated_signal
        self.sig.idx = self.sim_info.sim_buffer


    def propagate(self, num_samples):
        """Propagates signals from their sources to the microphones. 

        Generates signals from the sources, updates the RIRs if the sources or microphones are dynamic, and
        then filters the source signals through the RIRs. 
        
        Parameters
        ----------
        num_samples : int
            Number of samples to propagate.
        
        Notes
        -----
        The mic_signals are calculated for the indices self.sig.idx (inclusive) to self.sig.idx+num_samples (exclusive)
        """

        i = self.sig.idx

        for src in self.arrays.free_sources():
            self.sig[src.name][..., i:i+num_samples] = src.get_samples(num_samples)

        for src, mic in self.arrays.mic_src_combos():
            if src.dynamic or mic.dynamic:
                self.path_filters[src.name][mic.name].update_ir(self.arrays.paths[src.name][mic.name])
        
        for src, mic in self.arrays.mic_src_combos():
            propagated_signal = self.path_filters[src.name][mic.name].process(
                    self.sig[src.name][...,i:i+num_samples])
            self.sig[mic.name][...,i:i+num_samples] += propagated_signal
            if self.sim_info.save_source_contributions:
                self.sig[src.name+"~"+mic.name][...,i:i+num_samples] = propagated_signal
        #self.sig.idx += num_samples


import numpy as np
import matplotlib.pyplot as plt
import soundfile as sf
import json


import aspsim.fileutilities as fu
import aspsim.diagnostics.soundfieldplot as sfplot




def delete_earlier_tikz_plot(folder, name):
    currentIdx = fu.find_index_in_name(name)
    if currentIdx is None:
        return

    earlierFiles = fu.find_all_earlier_files(folder, name, currentIdx)
    for f in earlierFiles:
        if f.is_dir():
            for plotFile in f.iterdir():
                # assert(plotFile.stem.startswith(startName))
                try:
                    if plotFile.suffix == ".pdf":
                        plotFile.rename(
                            folder.joinpath(plotFile.stem + plotFile.suffix)
                        )
                    else:
                        assert plotFile.suffix == ".tsv" or plotFile.suffix == ".tex"
                        plotFile.unlink()
                except PermissionError:
                    pass
            try:
                f.rmdir()
            except PermissionError:
                pass

def tikzplotlib_fix_ncols(obj):
    """
    workaround for matplotlib 3.6 renamed legend's _ncol to _ncols, which breaks tikzplotlib
    """
    if hasattr(obj, "_ncols"):
        obj._ncol = obj._ncols
    for child in obj.get_children():
        tikzplotlib_fix_ncols(child)


def output_plot(print_method, folder, name="", keep_only_latest_tikz=False):
    if print_method == "show":
        plt.show()
    elif print_method == "tikz":
        import tikzplotlib
        if folder is not None:
            import tikzplotlib
            nested_folder = folder.joinpath(name)
            try:
                nested_folder.mkdir()
            except FileExistsError:
                pass

            fig = plt.gcf()
            tikzplotlib_fix_ncols(fig)
            tikzplotlib.save(
                str(nested_folder.joinpath(f"{name}.tex")),
                externalize_tables=True,
                #tex_relative_path_to_data="../figs/" + name + "/",
                float_format=".8g",
            )
            plt.savefig(
                str(nested_folder.joinpath(name + ".pdf")),
                dpi=300,
                facecolor="w",
                edgecolor="w",
                orientation="portrait",
                format="pdf",
                transparent=True,
                bbox_inches=None,
                pad_inches=0.2,
            )
            if keep_only_latest_tikz:
                delete_earlier_tikz_plot(folder, name)
    elif print_method == "pdf":
        if folder is not None:
            plt.savefig(
                str(folder.joinpath(name + ".pdf")),
                dpi=300,
                facecolor="w",
                edgecolor="w",
                orientation="portrait",
                format="pdf",
                transparent=True,
                #bbox_inches=None,
                bbox_inches="tight",
                pad_inches=0.2,
            )
    elif print_method == "svg":
        if folder is not None:
            plt.savefig(
                str(folder.joinpath(name + ".svg")),
                dpi=300,
                format="svg",
                transparent=True,
                bbox_inches="tight",
                pad_inches=0.2,
            )
    elif print_method == "none":
        pass
    else:
        raise ValueError("plot save method not recognized")
    plt.close("all")



def set_basic_plot_look(ax):
    ax.grid(True)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

def remove_axes_and_labels(ax):
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)

def plot_3d_in_2d(ax, posData, symbol="x", name=""):
    uniqueZValues = np.unique(posData[:,2].round(decimals=4))
    
    # if len(uniqueZValues) == 1:
    #     alpha = np.array([1])
    # else:
    #     alpha = np.linspace(0.4, 1, len(uniqueZValues))

    for i, zVal in enumerate(uniqueZValues):
        idx = np.where(np.around(posData[:,2],4) == zVal)[0]

        ax.plot(posData[idx,0], posData[idx,1], symbol, label=f"{name}: z = {zVal}")










def function_of_time_plot(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    #fig, ax = plt.subplots(1, 1, figsize=(14, 8))
    fig, ax = plt.subplots(1, 1, figsize=(7, 4))
    fig.tight_layout(pad=4)

    for proc_name, diag in diags.items():
        output, time_indices = diag.get_processed_output(time_idx, preprocess)

        num_channels = output.shape[0]

        if "label_suffix_channel" in diag.plot_data:
            labels = ["_".join((proc_name, suf)) for suf in diag.plot_data["label_suffix_channel"]]
            assert len(labels) == num_channels
        else:
            labels = [proc_name for _ in range(num_channels)]
        ax = plot_multiple_channels(ax, time_indices, output, labels)

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.grid(True)
    legend_without_duplicates(ax, "upper right")
    #ax.legend(loc="upper right")

    ax.set_xlabel(diag.plot_data["xlabel"])
    ax.set_ylabel(diag.plot_data["ylabel"])
    ax.set_title(diag.plot_data["title"] + " - " + name)
    output_plot(print_method, folder, name + "_" + str(time_idx))

def plot_multiple_channels(ax, time_idx, signal, labels):
    if signal.shape[-1] < 10:
        marker = "x"
    else:
        marker = ""
    for i, label in enumerate(labels):
        ax.plot(
                np.atleast_2d(time_idx).T,
                signal[i:i+1,:].T,
                alpha=0.8,
                label=label,
                marker=marker,
            )
    return ax

def legend_without_duplicates(ax, loc):
    handles, labels = ax.get_legend_handles_labels()
    newLabels, newHandles = [], []
    for handle, label in zip(handles, labels):
      if label not in newLabels:
        newLabels.append(label.replace("_", "\_"))
        newHandles.append(handle)

    ax.legend(newHandles, newLabels, loc=loc)


def savenpz(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    """Keeps only the latest save.
    Assumes that the data in previous
    saves is present in the current data"""
    outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}
    outputs = {key: val[0] if isinstance(val, tuple) else val for key, val in outputs.items()} 
    # only takes output even if time indices are provided
  
    #flatOutputs = util.flattenDict(outputs, sep="~")
    np.savez_compressed(folder.joinpath(f"{name}_{time_idx}"), **outputs)

    if list(diags.values())[0].keep_only_last_export:
        earlierFiles = fu.find_all_earlier_files(folder, name, time_idx, name_includes_idx=False)
        for f in earlierFiles:
            if f.suffix == ".npz":
                f.unlink()


def txt(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}
    outputs = {key: val[0] if isinstance(val, tuple) else val for key, val in outputs.items()} 
    # only takes output even if time indices are provided
    summary_val = {key : np.mean(val) for key, val in outputs.items()}
    
    with open(folder.joinpath(f"{name}_{time_idx}.json"), "w") as f:
        json.dump(summary_val, f)
    #np.savez_compressed(folder.joinpath(f"{name}_{time_idx}"), **outputs)


def soundfield(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    #outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}

    num_proc = len(diags)
    fig, axes = plt.subplots(1,num_proc, figsize=(num_proc*5, 5))
    if num_proc == 1:
        axes = [axes]

    sf_all = {proc_name : diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}

    max_val = np.max([np.max(sf) for sf in sf_all.values()])
    min_val = np.min([np.min(sf) for sf in sf_all.values()])

    for (proc_name, diag), sf, ax in zip(diags.items(), sf_all.values(), axes):
        
        pos = diag.pos_mic

        pos_sorted, sf_sorted = sfplot.sort_for_imshow(pos, sf[:,None])
        sfplot.sf_plot(ax, pos_sorted, sf_sorted, proc_name, vminmax=(min_val, max_val))

        if diag.plot_arrays is not None:
            for plot_ar in diag.plot_arrays:
                plot_ar.plot(ax)
        legend_without_duplicates(ax, "upper right")

    output_plot(print_method, folder, f"{name}_{time_idx}")

    #print("a plot would be generated at timeIdx: ", timeIdx, "for diagnostic: ", name)
	


def plot_ir(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    num_sets = 0
    for algo_name, diag in diags.items():
        for ir_set in diag.get_output():
            num_irs = ir_set.shape[0]
            num_sets += 1
    num_algo = len(diags)

    #fig, axes = plt.subplots(numSets, numIRs, figsize=(numSets*4, numIRs*4))
    fig, axes = plt.subplots(num_algo, num_irs, figsize=(num_irs*6, num_algo*6))
    if num_algo == 1:
        axes = np.expand_dims(axes, 0)
    if num_irs == 1:
        axes = np.expand_dims(axes, -1)
    #axes = np.atleast_2d(axes)
    #if not isinstance(axes, tuple):
    #    axes = [[axes]]
    #elif not isinstance(axes[0], tuple):
    #    axes = [axes]

    for row, (algo_name, diag) in enumerate(diags.items()):
        output = diag.get_processed_output(time_idx, preprocess)
        for set_idx, ir_set in enumerate(output):
            for col in range(ir_set.shape[0]):
                #axes[row, col].plot(irSet[col,:], label=f"{algoName} {metadata['label'][setIdx]}", alpha=0.6)
                axes[row, col].plot(ir_set[col,:], label=f"{algo_name}", alpha=0.6)
                #axes[row, col].set_xlabel(diag.plot_data["xlabel"])
                #axes[row, col].set_ylabel(diag.plot_data["ylabel"])

    for ax in axes.flatten():
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.grid(True)
        legend_without_duplicates(ax, "upper right")
        #ax.legend(loc="upper right")
        
    output_plot(print_method, folder, f"{name}_{time_idx}")

def matshow(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}
    
    num_proc = len(diags)
    fig, axes = plt.subplots(1,num_proc, figsize=(num_proc*5, 5))
    if num_proc == 1:
        axes = [axes]

    for ax, (proc_name, output) in zip(axes, outputs.items()):
        clr = ax.matshow(output)
        fig.colorbar(clr, ax=ax, orientation='vertical')
        ax.set_title(f"{name} - {proc_name}")
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
    output_plot(print_method, folder, f"{name}_{time_idx}")


def create_audio_files(name, diags, time_idx, folder, preprocess, print_method=None):
    outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}
    outputs = {key: val[0] if isinstance(val, tuple) else val for key, val in outputs.items()} 
    # only takes the signal even if time indices are provided

    samplerate = diags[list(diags.keys())[0]].sim_info.samplerate

    #max_val = np.NINF
    #last_time_idx = np.inf
    #for (proc_name, signal) in outputs.items():
    #        max_val = np.max((max_val, np.max(np.abs(signal[~np.isnan(signal)]))))
    #        last_time_idx = int(np.min((last_time_idx, np.max(np.where(~np.isnan(signal))))))

    #start_idx = np.max((last_time_idx-metadata["maxlength"], 0))

    for proc_name, signal in outputs.items():
        file_name = "_".join((name, proc_name, str(time_idx)))
        file_path = folder.joinpath(f"{file_name}.wav")


        # for channel_idx in range(signal.shape[0]):
        #     if signal.shape[0] > 1:
        #         fileName = "_".join((name, proc_name, audio_name, str(channel_idx), str(time_idx)))
        #     else:
        #         fileName = "_".join((name, proc_name, audio_name, str(time_idx)))
        #     file_path = folder.joinpath(fileName + ".wav")

        signal_to_write = signal / np.max(np.abs(signal))
        

        ramp_length = min(int(0.2*samplerate), int(0.1 * signal_to_write.shape[-1]))
        if ramp_length > 0:
            ramp = np.linspace(0,1,ramp_length)
            signal_to_write[:,:ramp_length] *= ramp
            signal_to_write[:,-ramp_length:] *= (1-ramp)

        sf.write(str(file_path), signal_to_write.T, samplerate)




def spectrum_plot(name, diags, time_idx, folder, preprocess, print_method="pdf"):
    #outputs = {proc_name: diag.get_processed_output(time_idx, preprocess) for proc_name, diag in diags.items()}
    #outputs = {key: val[0] if isinstance(val, tuple) else val for key, val in outputs.items()}

    fig, ax = plt.subplots(1, 1, figsize=(14, 8))
    fig.tight_layout(pad=4)

    for proc_name, diag in diags.items():
        output = diag.get_processed_output(time_idx, preprocess)
        num_channels = output.shape[0]
        ax.plot(
                output,
                alpha=0.8,
                label=proc_name,
            )

    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.grid(True)
    legend_without_duplicates(ax, "upper right")
        #ax.legend(loc="upper right")

        #ax.set_xlabel(diag.plot_data["xlabel"])
        #ax.set_ylabel(diag.plot_data["ylabel"])
        #ax.set_title(diag.plot_data["title"] + " - " + name)
    output_plot(print_method, folder, name + "_" + str(time_idx))




import numpy as np
import hypothesis as hyp
import hypothesis.strategies as st
import pytest

from aspsim.simulator import SimulatorSetup
import aspsim.diagnostics.diagnostics as dia
import aspsim.signal.sources as sources
import aspsim.room.trajectory as tr
import aspcore.filter as fc


import aspsim.configutil as cu

def _default_sim_info():

    sim_info = cu.load_default_config()
    sim_info.tot_samples = 20
    sim_info.sim_buffer = 20
    sim_info.export_frequency = 20
    sim_info.sim_chunk_size = 20
    sim_info.max_room_ir_length = 8

    sim_info.start_sources_before_0 = False
    return sim_info


@pytest.fixture(scope="session")
def fig_folder(tmp_path_factory):
    return tmp_path_factory.mktemp("figs")


def _setup_ism(fig_folder, samplerate):
    setup = SimulatorSetup(fig_folder)
    setup.sim_info.samplerate = samplerate
    setup.sim_info.tot_samples =  samplerate
    setup.sim_info.sim_chunk_size = 2*samplerate
    setup.sim_info.sim_buffer = samplerate
    setup.sim_info.export_frequency =  samplerate
    setup.sim_info.save_source_contributions = False
    setup.sim_info.randomized_ism = False

    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 5, 5]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.rt60 = 0.25
    setup.sim_info.max_room_ir_length = samplerate // 2
    return setup


def test_unmoving_trajectory_same_as_static(fig_folder):
    sr = 500
    sim_setup = _setup_ism(fig_folder, sr)
    def zero_pos_func(time):
        return np.zeros((1,3))

    mic_traj = tr.Trajectory(zero_pos_func)
    sim_setup.add_mics("mic", np.zeros((1,3)))
    sim_setup.add_mics("trajectory", mic_traj)
    sim_setup.add_free_source("source", np.array([[0, -1, -1]]), sources.WhiteNoiseSource(1, 1, np.random.default_rng(1)))

    sim = sim_setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.diag.add_diagnostic("trajectory", dia.RecordSignal("trajectory", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()

    sig_mic = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]
    
    sig_traj = np.load(sim.folder_path.joinpath(f"trajectory_{sim.sim_info.tot_samples}.npz"))["trajectory"]
    assert np.allclose(sig_mic, sig_traj)




def test_moving_microphone_equals_moving_source(fig_folder):
    """ A moving microphone should give the same output as a moving source, if the
    positions are switched. This is an extension of the reciprocity test, but with
    moving sources. 

    Actually, I don't think this is true...

    Currently gives an sample-wise error of around 1e-5, which is unclear if it is 
    within tolerance for numerical errors or not. Probably a valid result, just that
    the test is actually not true.
    """
    sr = 500
    rng = np.random.default_rng()

    pos_stationary = np.zeros((1,3))
    pos_moving = tr.LinearTrajectory([[1,1,1], [0,1,0], [1,0,1]], 1, sr)

    sim_setup = _setup_ism(fig_folder, sr)
    src_sig = rng.random(size=(1, sim_setup.sim_info.tot_samples*2))

    
    sim_setup.add_mics("mic", pos_stationary)
    sim_setup.add_free_source("src", pos_moving, sources.Sequence(src_sig))
    sim = sim_setup.create_simulator()
    sim.diag.add_diagnostic("mic1", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()
    sig1 = np.load(sim.folder_path.joinpath(f"mic1_{sim.sim_info.tot_samples}.npz"))["mic1"]

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos_moving)
    sim_setup.add_free_source("src", pos_stationary, sources.Sequence(src_sig))
    sim = sim_setup.create_simulator()
    sim.diag.add_diagnostic("mic2", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()

    sig2 = np.load(sim.folder_path.joinpath(f"mic2_{sim.sim_info.tot_samples}.npz"))["mic2"]
    assert np.allclose(sig1, sig2)

def test_moving_microphone_is_using_expected_positions(fig_folder):
    rng = np.random.default_rng()

    setup = SimulatorSetup(fig_folder)
    setup.sim_info = _default_sim_info()

    pos_src = np.zeros((1,3))
    pos_mic = tr.LinearTrajectory([[1,1,1], [0,1,0], [1,0,1]], 1, setup.sim_info.samplerate)

    setup.add_mics("mic", pos_mic)
    setup.add_free_source("src", pos_src, sources.WhiteNoiseSource(1,1, rng))
    sim = setup.create_simulator()
    sim.run_simulation()
    pos_all = np.array(sim.arrays["mic"].pos_all)
    time_all = np.array(sim.arrays["mic"].time_all)

    pos_compare = np.array([pos_mic.current_pos(t) for t in range(setup.sim_info.tot_samples)])
    assert np.allclose(pos_all[:setup.sim_info.tot_samples], pos_compare)
   


def test_moving_microphone_has_same_rirs_as_stationary_microphones_on_trajectory(fig_folder):
    rng = np.random.default_rng()
    sr = 500

    setup = _setup_ism(fig_folder, sr)
    src_sig = rng.random(size=(1, setup.sim_info.tot_samples*2))

    pos_src = np.zeros((1,3))
    traj = tr.LinearTrajectory([[1,1,1], [0,1,0], [1,0,1]], 1, setup.sim_info.samplerate)
    all_pos = np.array([traj.current_pos(t) for t in range(setup.sim_info.tot_samples)])[:,0,:]

    setup.add_mics("traj", traj)
    setup.add_mics("mic", all_pos)
    setup.add_free_source("src", pos_src, sources.Sequence(src_sig))
    sim = setup.create_simulator()
    
    sim.run_simulation()

    rir1 = np.array(sim.arrays._rir_dynamic_all)[:setup.sim_info.tot_samples,0,0,:]
    rir2 = sim.arrays.paths["src"]["mic"][0,:,:]
    #sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, num_channels = all_pos.shape[0], export_func="npz"))
    #sim.diag.add_diagnostic("traj_rir", dia.RecordState("traj", sim.sim_info, num_channels = 1, export_func="npz"))
    #sim.run_simulation()


    #sig_mic = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]
    #sig_traj_reconstruct = np.array([sig_mic[i,i] for i in range(sig_mic.shape[0])])[None,:]
    #sig_traj = np.load(sim.folder_path.joinpath(f"traj_{sim.sim_info.tot_samples}.npz"))["traj"]

    assert np.allclose(rir1, rir2)



def test_moving_microphone_gives_same_output_as_pointwise_stationary_convolutions(fig_folder):
    """The equivalence is currently for the range (-1, tot_samples-1), which is not the 
    intended behaviour. The sim should be changed to correctly give equivalence for (0, tot_samples)
    """
    rng = np.random.default_rng()
    sr = 500

    setup = _setup_ism(fig_folder, sr)
    src_sig = rng.random(size=(1, setup.sim_info.tot_samples*2))

    pos_src = np.zeros((1,3))
    traj = tr.LinearTrajectory([[1,1,1], [0,1,0], [1,0,1]], 1, setup.sim_info.samplerate)
    all_pos = np.array([traj.current_pos(t) for t in range(-1, setup.sim_info.tot_samples-1)])[:,0,:]
    all_pos[0,:] += 2 # position at index 0 does not matter
    all_pos[1,:] += 2 # position at index 1 does not matter

    setup.add_mics("traj", traj)
    setup.add_mics("mic", all_pos)
    setup.add_free_source("src", pos_src, sources.Sequence(src_sig))
    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, num_channels = all_pos.shape[0], export_func="npz"))
    sim.diag.add_diagnostic("traj", dia.RecordSignal("traj", sim.sim_info, num_channels = 1, export_func="npz"))
    sim.run_simulation()


    sig_mic = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]
    sig_traj_reconstruct = np.array([sig_mic[i,i] for i in range(sig_mic.shape[0])])[None,:]
    sig_traj = np.load(sim.folder_path.joinpath(f"traj_{sim.sim_info.tot_samples}.npz"))["traj"]

    assert np.allclose(sig_traj, sig_traj_reconstruct)




import numpy as np


def cart2spherical(cart_coord):
    """Transforms the provided cartesian coordinates to spherical coordinates

    Parameters
    ----------
    cart_coord : ndarray of shape (num_points, 3)

    Returns
    -------
    r : ndarray of shape (num_points, 1)
        radius of each point
    angle : ndarray of shape (num_points, 2)
        angle[:,0] is theta, the angle in the xy plane, where 0 is x direction, pi/2 is y direction
        angle[:,1] is phi, the zenith angle, where 0 is z direction, pi is negative z direction
    """
    r = np.linalg.norm(cart_coord, axis=1)
    r_xy = np.linalg.norm(cart_coord[:,:2], axis=1)

    theta = np.arctan2(cart_coord[:,1], cart_coord[:,0])
    phi = np.arctan2(r_xy, cart_coord[:,2])
    angle = np.concatenate((theta[:,None], phi[:,None]), axis=1)
    return (r, angle)

def spherical2cart(r, angle):
    """Transforms the provided spherical coordinates to cartesian coordinates
    
    Parameters
    ----------
    r : ndarray of shape (num_points, 1) or (num_points,)
        radius of each point
    angle : ndarray of shape (num_points, 2)
        the angles in radians
        angle[:,0] is theta, the angle in the xy plane, where 0 is x direction, pi/2 is y direction
        angle[:,1] is phi, the zenith angle, where 0 is z direction, pi is negative z direction
    
    Returns
    -------
    cart_coord : ndarray of shape (num_points, 3)
        the cartesian coordinates
    """
    num_points = r.shape[0]
    cart_coord = np.zeros((num_points,3))
    cart_coord[:,0] = np.squeeze(r) * np.cos(angle[:,0]) * np.sin(angle[:,1])
    cart_coord[:,1] = np.squeeze(r) * np.sin(angle[:,0]) * np.sin(angle[:,1])
    cart_coord[:,2] = np.squeeze(r) * np.cos(angle[:,1])
    return cart_coord

def cart2pol(x, y):
    """Transforms the provided cartesian coordinates to polar coordinates

    Parameters
    ----------
    x : ndarray of shape (num_points,)
        x-coordinate of each point
    y : ndarray of shape (num_points,)
        y-coordinate of each point

    Returns
    -------
    r : ndarray of shape (num_points,)
        radius of each point
    angle : ndarray of shape (num_points,)
        the angles in radians
    """
    r = np.hypot(x, y)
    angle = np.arctan2(y, x)
    return (r, angle)

def pol2cart(r, angle):
    """Transforms the provided polar coordinates to cartesian coordinates

    Parameters
    ----------
    r : ndarray of shape (num_points,)
        radius of each point
    angle : ndarray of shape (num_points,)
        the angles in radians

    Returns
    -------
    x : ndarray of shape (num_points,)
        x-coordinate of each point
    y : ndarray of shape (num_points,)
        y-coordinate of each point
    """
    x = r * np.cos(angle)
    y = r * np.sin(angle)
    return (x, y)


def equiangular_circle(num_points, radius, start_angle=0, z=None, rng=None):
    """Generates equiangularly spaced points on a circle

    Parameters
    ----------
    num_points : int
        number of points on the circle
    radius : int or 2-tuple
        radius of the circle
        if a tuple is suppled, as (inner_radius, outer_radius)
        each point will be placed randomly (uniform distribution)
        between the inner and outer radius 
    start_angle : float
        the angle in radians of the first point
    z : None or float
        if z is supplied, the all points are given the same z-coordinate
        according to the value of the argument
    rng : None or numpy Generator object
        can be supplied to control the seed of the random number
        generator
    

    Returns
    -------
    coordinates : ndarray of shape (num_points, 2) or (num_points, 3)
        if argument z was supplied
    """
    if rng is None:
        rng = np.random.default_rng()
    
    if isinstance(radius, (int, float)):
        radius = (radius, radius)

    angle_step = 2 * np.pi / num_points

    angles = start_angle + np.arange(num_points) * angle_step
    angles = np.mod(angles, 2 * np.pi)
    radia = rng.uniform(radius[0], radius[1], size=num_points)
    [x, y] = pol2cart(radia, angles)

    if z is not None:
        pos = np.zeros((num_points, 3))
        pos[:, 2] = z
    else:
        pos = np.zeros((num_points, 2))
    pos[:, 0:2] = np.stack((x, y)).T
    return pos



def equidistant_rectangle(num_points, side_lengths, extra_side_lengths = 0, offset= 0.5, z = None): 
    """Points are spaced evenly along the edge of a rectangle
    
    Parameters
    ----------
    num_points : int
        the number of total points to be returned. These will be distributed as 
        evenly as possible
    side_lengths : array_like of shape (2,)
        the side lengths in x and y coordinates. The corners of the rectangle will be at
        (-sl[0]/2, -sl[1]/2), (-sl[0]/2, sl[1]/2), (sl[0]/2, -sl[1]/2), (sl[0]/2, sl[1]/2)
        where sl is shorthand for side_lengths
    extra_side_lengths : float larger or equal to 0, or a 2-tuple of such values
        if set to a non-zero value, every other point will be placed further away from the center
        as if the points are placed on two rectangles, the larger with side lengths of
        (side_lengths[0] + extra_side_lengths[0], side_lengths[1] + extra_side_lengths[1])
    offset : float within [0,1]
        The offset from the first corner to the first microphones. If offset is 0, the first
        microphone is placed in the corner. If offset is 0.5, the first microphone is placed
        as close to the middle of its 'segment' as possible
    z : float
        The z coordinate, which will be the same for all points in the rectangle. If not provided, 
        a 2D shape will be returned

    Returns
    -------
    rectangle_points : ndarray of shape (num_points, 3) or (num_points, 2)
        The coordinates of the points in the rectangle
        the shape is (num_points, 3) if z is provided, and (num_points, 2) otherwise. 
    """
    side_lengths = np.array(side_lengths)
    assert side_lengths.shape == (2,)
    assert np.all(side_lengths > 0)
    if not isinstance(extra_side_lengths, (list, tuple, np.ndarray)):
        extra_side_lengths = [extra_side_lengths, extra_side_lengths]
    assert 0 <= offset <= 1

    circumference = 2 * np.sum(side_lengths)
    pos_on_circumference = np.linspace(0, circumference, num_points, endpoint = False)
    mic_distance = circumference / num_points
    pos_on_circumference += offset * mic_distance

    x_extent = np.array([-side_lengths[0], side_lengths[0]]) / 2
    y_extent = np.array([-side_lengths[1], side_lengths[1]]) / 2
    fixed_side = [y_extent[1], x_extent[1], y_extent[0], x_extent[0]] #side with index i has a fixed coordinate for x or y. this is that coordinate

    points = np.zeros((num_points, 2))
    breakpoints = np.cumsum(np.array([0, side_lengths[0], side_lengths[1], side_lengths[0], side_lengths[1]]))
    idxs = [np.logical_and(breakpoints[i] <= pos_on_circumference, pos_on_circumference < breakpoints[i+1]) for i in range(4)]
    
    points = []
    pos_side = pos_on_circumference[idxs[0]] - breakpoints[0] - x_extent[1]
    pos_fixed = np.full(num_points, fixed_side[0])
    pos_fixed[np.arange(num_points) % 2 == 1] += extra_side_lengths[1] / 2
    pos_fixed = pos_fixed[idxs[0]]
    points.append(np.stack([pos_side, pos_fixed], axis=-1))

    pos_side = pos_on_circumference[idxs[1]] - breakpoints[1] - y_extent[1]
    pos_side = -pos_side
    pos_fixed = np.full(num_points, fixed_side[1])
    pos_fixed[np.arange(num_points) % 2 == 1] += extra_side_lengths[0] / 2
    pos_fixed = pos_fixed[idxs[1]]
    points.append(np.stack([pos_fixed, pos_side], axis=-1))

    pos_side = np.flip(pos_on_circumference[idxs[2]]) - breakpoints[2] - x_extent[1]
    pos_side = -pos_side
    pos_fixed = np.full(num_points, fixed_side[2])
    pos_fixed[np.arange(num_points) % 2 == 1] -= extra_side_lengths[1] / 2
    pos_fixed = pos_fixed[idxs[2]]
    points.append(np.stack([pos_side, pos_fixed], axis=-1))

    pos_side = pos_on_circumference[idxs[3]] - breakpoints[3] - y_extent[1]
    pos_fixed = np.full(num_points, fixed_side[3])
    pos_fixed[np.arange(num_points) % 2 == 1] -= extra_side_lengths[0] / 2
    pos_fixed = pos_fixed[idxs[3]]
    points.append(np.stack([pos_fixed, pos_side], axis=-1))

    points = np.concatenate(points, axis=0)
    if z is not None:
        points = np.concatenate((points, np.full((points.shape[0], 1), z)), axis=-1)
    return points







def concentrical_circles(
    num_points, num_circles, radius, z_distance=None, angle_offset="distribute"
):
    a = num_circles // 2
    if z_distance is not None:
        zValues = [i * z_distance for i in range(-a, -a + num_circles)]
        if num_circles % 2 == 0:
            zValues = [zVal + z_distance / 2 for zVal in zValues]
    else:
        zValues = [None for i in range(num_circles)]

    pointsPerCircle = [num_points // num_circles for _ in range(num_circles)]
    for i in range(num_points - num_circles * (num_points // num_circles)):
        pointsPerCircle[i] += 1

    if angle_offset == "random":
        startAngles = np.random.rand(num_circles) * np.pi * 2
    elif angle_offset == "same":
        startAngles = np.zeros(num_circles)
    elif angle_offset == "almostSame":
        angleSection = 2 * np.pi / pointsPerCircle[0]
        startAngles = np.random.rand(num_circles) * (angleSection / 10)
    elif angle_offset == "distribute":
        angleSection = 2 * np.pi / pointsPerCircle[0]
        startAngles = np.arange(num_circles) * angleSection / num_circles

    coords = [
        equiangular_circle(circlePoints, radius, angle, zVal)
        for i, (circlePoints, angle, zVal) in enumerate(
            zip(pointsPerCircle, startAngles, zValues)
        )
    ]

    coords = np.concatenate(coords)
    return coords





def uniform_cylinder(num_points, radius, height):
    numPlanes = 4
    zVals = np.linspace(-height / 2, height / 2, numPlanes + 2)
    zVals = zVals[1:-1]

    pointsPerPlane = num_points // numPlanes
    allPoints = np.zeros((pointsPerPlane * numPlanes, 3))

    for n in range(numPlanes):
        xyPoints = sunflower_pattern(
            pointsPerPlane, radius, np.random.rand() * 2 * np.pi
        )
        allPoints[n * pointsPerPlane : (n + 1) * pointsPerPlane :, 0:2] = xyPoints
        allPoints[n * pointsPerPlane : (n + 1) * pointsPerPlane, 2] = zVals[n]
    return allPoints



def sunflower_pattern(N, radius, offset_angle=0):
    """ translated from user3717023's MATLAB code from stackoverflow
        could be updated using the method in this paper 
        'A better way to construct the sunflower head'"""
    phisq = np.square((np.sqrt(5) + 1) / 2)
    # golden ratio
    k = np.arange(1, N + 1)
    r = radius * np.sqrt(k - (1 / 2)) / np.sqrt(N - 1 / 2)
    theta = k * 2 * np.pi / phisq
    [x, y] = pol2cart(r, theta)
    return np.stack((x, y)).T


def uniform_disc(point_distance, radius, z_axis=None):
    lim = (-radius, radius)
    numPoints = int(2 * radius / point_distance)
    x = np.linspace(lim[0], lim[1], numPoints)
    y = np.linspace(lim[0], lim[1], numPoints)
    [xGrid, yGrid] = np.meshgrid(x, y)
    xGrid = xGrid.flatten()
    yGrid = yGrid.flatten()

    coords = np.vstack((xGrid, yGrid))
    dist = np.linalg.norm(coords, axis=0)
    idxs2 = dist <= radius
    coordsCircle = coords[:, idxs2].T

    if z_axis is not None:
        coordsCircle = np.concatenate(
            (coordsCircle, np.full((coordsCircle.shape[0], 1), z_axis)), axis=-1
        )
    return coordsCircle


def uniform_filled_rectangle(num_points, lim=(-2.4, 2.4), z_axis=None):
    pointsPerAxis = int(np.sqrt(num_points))
    assert np.isclose(pointsPerAxis ** 2, num_points)
    if len(lim) == 2:
        x = np.linspace(lim[0], lim[1], pointsPerAxis)
        y = np.linspace(lim[0], lim[1], pointsPerAxis)
    elif len(lim) == 4:
        x = np.linspace(lim[0], lim[2], pointsPerAxis)
        y = np.linspace(lim[1], lim[3], pointsPerAxis)

    [xGrid, yGrid] = np.meshgrid(x, y)
    evalPoints = np.vstack((xGrid.flatten(), yGrid.flatten())).T

    if z_axis is not None:
        evalPoints = np.concatenate(
            (evalPoints, np.full((pointsPerAxis ** 2, 1), z_axis)), axis=-1
        )
    return evalPoints

def uniform_filled_cuboid(num_points, dims, z_num_points=4):
    pointsPerAxis = int(np.sqrt(num_points / z_num_points))
    assert np.isclose(pointsPerAxis ** 2 * z_num_points, num_points)
    x = np.linspace(-dims[0] / 2, dims[0] / 2, 2 * pointsPerAxis + 1)[1::2]
    y = np.linspace(-dims[1] / 2, dims[1] / 2, 2 * pointsPerAxis + 1)[1::2]
    z = np.linspace(-dims[2] / 2, dims[2] / 2, 2 * z_num_points + 1)[1::2]
    [xGrid, yGrid, zGrid] = np.meshgrid(x, y, z)
    evalPoints = np.vstack((xGrid.flatten(), yGrid.flatten(), zGrid.flatten())).T

    return evalPoints


def four_equidistant_rectangles(
    num_points, side_length, side_offset, z_low, z_high, offset="distributed"
):
    if offset != "distributed":
        raise NotImplementedError
    points = np.zeros((num_points, 3))
    points[:, 0:2] = equidistant_rectangle(num_points, (side_length, side_length))
    # points[0::2,2] = zLow
    # points[1::2,2] = zHigh

    # idxSet = np.sort(np.concatenate((np.arange(numPoints)[2::4], np.arange(numPoints)[3::4])))
    # for i in idxSet:
    #     if np.isclose(points[i,0], sideLength/2):
    #         points[i,0] += sideOffset
    #     elif np.isclose(points[i,0], -sideLength/2):
    #         points[i,0] -= sideOffset
    #     elif np.isclose(points[i,1], sideLength/2):
    #         points[i,1] += sideOffset
    #     elif np.isclose(points[i,1], -sideLength/2):
    #         points[i,1] -= sideOffset
    #     else:
    #         raise ValueError
    idxSet = np.sort(
        np.concatenate((np.arange(num_points)[2::4], np.arange(num_points)[3::4]))
    )
    points[:, 2] = z_low
    points[idxSet, 2] = z_high
    # points[]
    # points[0::2,2] = zLow
    # points[1::2,2] = zHigh

    for i in np.arange(num_points)[1::2]:
        if np.isclose(points[i, 0], side_length / 2):
            points[i, 0] += side_offset
        elif np.isclose(points[i, 0], -side_length / 2):
            points[i, 0] -= side_offset
        elif np.isclose(points[i, 1], side_length / 2):
            points[i, 1] += side_offset
        elif np.isclose(points[i, 1], -side_length / 2):
            points[i, 1] -= side_offset
        else:
            raise ValueError
    return points


def stacked_equidistant_rectangles(
    num_points, num_rect, dims, z_distance, offset="distributed"
):
    a = num_rect // 2
    zValues = [i * z_distance for i in range(-a, -a + num_rect)]
    if num_rect % 2 == 0:
        zValues = [zVal + z_distance / 2 for zVal in zValues]

    pointsPerRect = [num_points // num_rect for _ in range(num_rect)]
    for i in range(num_points - num_rect * (num_points // num_rect)):
        pointsPerRect[i] += 1

    offsets = np.linspace(0, 1, 2 * num_rect + 1)[1::2]

    points = np.zeros((num_points, 3))
    idxCount = 0
    for i in range(num_rect):
        points[idxCount : idxCount + pointsPerRect[i], 0:2] = equidistant_rectangle(
            pointsPerRect[i], dims, offset=offsets[i]
        )
        points[idxCount : idxCount + pointsPerRect[i], 2] = zValues[i]
        idxCount += pointsPerRect[i]
    return points


def equidistant_rectangle_old(num_points, dims, offset=0.5, z=None):
    if num_points == 0:
        if z is None:
            return np.zeros((0, 2))
        else:
            return np.zeros((0,3))
    totalLength = 2 * (dims[0] + dims[1])
    pointDist = totalLength / num_points

    points = np.zeros((num_points, 2))
    if num_points < 4:
        points = equidistant_rectangle(4, dims)
        #pointChoices = np.random.choice(4, numPoints, replace=False)
        #points = points[pointChoices, :]
        points = points[:num_points,:]
    else:
        lengths = [dims[0], dims[1], dims[0], dims[1]]
        xVal = [-dims[0] / 2, dims[0] / 2, dims[0] / 2, -dims[0] / 2]
        yVal = [-dims[1] / 2, -dims[1] / 2, dims[1] / 2, dims[1] / 2]

        startPos = pointDist * offset
        xFac = [1, 0, -1, 0]
        yFac = [0, 1, 0, -1]
        numCounter = 0

        for i in range(4):
            numAxisPoints = 1 + int((lengths[i] - startPos) / pointDist)
            axisPoints = startPos + np.arange(numAxisPoints) * pointDist
            distLeft = lengths[i] - axisPoints[-1]
            points[numCounter : numCounter + numAxisPoints, 0] = (
                xVal[i] + xFac[i] * axisPoints
            )
            points[numCounter : numCounter + numAxisPoints, 1] = (
                yVal[i] + yFac[i] * axisPoints
            )
            numCounter += numAxisPoints
            startPos = pointDist - distLeft
    if z is not None:
        points = np.concatenate((points, np.full((points.shape[0], 1), z)), axis=-1)
    return points


def equidistant_rectangle_for_fewer(num_points, dims):
    totalLength = 2 * (dims[0] + dims[1])
    pointDist = totalLength / num_points

    points = np.zeros((num_points, 2))
    if num_points == 1:
        points[0, 0] = np.random.rand() * dims[0] / 4 - dims[0] / 4
        points[0, 1] = np.random.choice([-1, 1]) * dims[1] / 2
    elif num_points == 2:
        points[:, 0] = np.random.rand(2) * dims[0] / 4 - dims[0] / 4
        points[0, 1] = dims[1] / 2
        points[1, 1] = dims[1] / 2
    elif num_points == 3:
        points[0:2, 0] = np.random.rand(2) * dims[0] / 4 - dims[0] / 4
        points[0, 1] = dims[1] / 2
        points[1, 1] = dims[1] / 2
        points[2, 0] = np.random.choice([-1, 1]) * dims[0] / 2
        points[2, 1] = np.random.rand() * dims[1] / 4 - dims[1] / 4
    else:
        lengths = [dims[0], dims[1], dims[0], dims[1]]
        xVal = [-dims[0] / 2, dims[0] / 2, dims[0] / 2, -dims[0] / 2]
        yVal = [-dims[1] / 2, -dims[1] / 2, dims[1] / 2, dims[1] / 2]
        startPos = np.random.rand() * pointDist
        xFac = [1, 0, -1, 0]
        yFac = [0, 1, 0, -1]
        numCounter = 0

        for i in range(np.min((4, num_points))):
            numAxisPoints = 1 + int((lengths[i] - startPos) / pointDist)
            axisPoints = startPos + np.arange(numAxisPoints) * pointDist
            distLeft = lengths[i] - axisPoints[-1]
            points[numCounter : numCounter + numAxisPoints, 0] = (
                xVal[i] + xFac[i] * axisPoints
            )
            points[numCounter : numCounter + numAxisPoints, 1] = (
                yVal[i] + yFac[i] * axisPoints
            )
            numCounter += numAxisPoints
            startPos = pointDist - distLeft

    return points


from pathlib import Path
import yaml
import dataclasses as dc


@dc.dataclass
class SimulatorInfo:
    tot_samples : int
    sim_buffer : int
    sim_chunk_size : int

    start_sources_before_0 : bool
    save_source_contributions : bool
    array_update_freq : int

    samplerate : int
    c : float
    spatial_dims : int
    reverb : str

    room_size : list[float]
    room_center : list[float]
    rt60 : float
    max_room_ir_length : int
    randomized_ism : bool
    extra_delay : int
    highpass_cutoff : float

    export_frequency : int
    plot_output : str
    output_smoothing : int
    auto_save_load : bool


    def __post_init__(self):
        # Should check here that sim_buffer is large enough that it wont cause errors
        assert len(self.room_size) == self.spatial_dims
        assert len(self.room_center) == self.spatial_dims

        assert self.reverb in ("none", "direct", "ism")


    def save_to_file(self, path):
        if path is not None:
            with open(path.joinpath("config.yaml"), "w") as f:
                yaml.dump(dc.asdict(self), f, sort_keys=False)

def load_from_file(path):
    if path.is_dir():
        path = path.joinpath("config.yaml")

    with open(path) as f:
        config = yaml.safe_load(f)
    return SimulatorInfo(**config)


def load_default_config():
    path = Path(__file__).parent.joinpath("config.yaml")
    return load_from_file(path)


def equal_audio(info1, info2, path_types):
    same_audio =  info1.samplerate == info2.samplerate and \
                info1.c == info2.c and \
                info1.spatial_dims == info2.spatial_dims and \
                info1.reverb == info2.reverb
    
    if any("ism" in src_path_types.values() for src_path_types in path_types.values()):
        same_audio = same_audio and \
            info1.room_size == info2.room_size and \
            info1.room_center == info2.room_center and \
            info1.rt60 == info2.rt60 and \
            info1.max_room_ir_length == info2.max_room_ir_length and \
            info1.randomized_ism == info2.randomized_ism and \
            info1.extra_delay == info2.extra_delay and \
            info1.highpass_cutoff == info2.highpass_cutoff
    return same_audio




import shutil
import json

import aspsim.fileutilities as futil
import aspsim.configutil as configutil
import aspsim.array as ar



def save_session(session_folder, sim_info, arrays, sim_metadata=None, extraprefix=""):
    session_path = futil.get_unique_folder_name("session_" + extraprefix, session_folder)

    session_path.mkdir()
    arrays.save_to_file(session_path)
    sim_info.save_to_file(session_path)
    if sim_metadata is not None:
        add_to_sim_metadata(session_path, sim_metadata)

def load_session(sessions_path, new_folder_path, chosen_sim_info, chosen_arrays):
    """sessionsPath refers to the folder where all sessions reside 
        This function will load a session matching the chosenConfig 
        and chosenArrays if it exists"""
    session_to_load = search_for_matching_session(sessions_path, chosen_sim_info, chosen_arrays)
    print("Loaded Session: ", str(session_to_load))
    loaded_arrays = ar.load_arrays(session_to_load)

    for fs_array in chosen_arrays.free_sources():
        loaded_arrays[fs_array.name].source = fs_array.source

    return loaded_arrays
    
def load_from_path(session_path_to_load, new_folder_path=None):
    loaded_arrays = ar.load_arrays(session_path_to_load)
    loaded_sim_info = configutil.load_from_file(session_path_to_load)

    if new_folder_path is not None:
        copy_sim_metadata(session_path_to_load, new_folder_path)
        loaded_sim_info.save_to_file(new_folder_path)
    return loaded_sim_info, loaded_arrays

def copy_sim_metadata(from_folder, to_folder):
    shutil.copy(
        from_folder.joinpath("metadata_sim.json"), to_folder.joinpath("metadata_sim.json")
    )


class MatchingSessionNotFoundError(ValueError): pass

def search_for_matching_session(sessions_path, chosen_sim_info, chosen_arrays):
    for dir_path in sessions_path.iterdir():
        if dir_path.is_dir():
            loaded_sim_info = configutil.load_from_file(dir_path)
            loaded_arrays = ar.load_arrays(dir_path)

            if configutil.equal_audio(chosen_sim_info, loaded_sim_info, chosen_arrays.path_type) and \
                ar.prototype_equals(chosen_arrays, loaded_arrays):
                return dir_path
    raise MatchingSessionNotFoundError("No matching saved sessions")



def add_to_sim_metadata(folder_path, dict_to_add):
    try:
        with open(folder_path.joinpath("metadata_sim.json"), "r") as f:
            old_data = json.load(f)
            tot_data = {**old_data, **dict_to_add}
    except FileNotFoundError:
        tot_data = dict_to_add
    with open(folder_path.joinpath("metadata_sim.json"), "w") as f:
        json.dump(tot_data, f, indent=4)


def write_processor_metadata(processors, folder_path):
    if folder_path is None:
        return
        
    file_name = "metadata_processor.json"
    tot_metadata = {}
    for proc in processors:
        tot_metadata[proc.name] = proc.metadata
    with open(folder_path.joinpath(file_name), "w") as f:
        json.dump(tot_metadata, f, indent=4)


import numpy as np
from abc import abstractmethod

import aspsim.diagnostics.plot as dplot


"""
===== DIAGNOSTICS OVERVIEW =====
All diagnostics should inherit from Diagnostic, but it is suggested to use the 
intermediate SignalDiagnostic, StateDiagnostic, and Instantdiagnostic. 

===== PROCESSOR.IDX =====
The processor.idx should be interpreted as the next (local, in reference to the internal signal buffer)
time index it will process. So if processor.idx == 1, then processor.sig['signame'][:,0] is a processed sample
that can be saved to a diagnostic, but processor.sig['signame'][:,1] is not. 

The processor.idx is increased after propagating all signals. During its .process method, the processing can
add source signals between idx:idx+block_size. These new signals are propagated to the microphones, and the 
index is increased by block_size. Directly after that, the diagnostics are saved. 

===== GLOBAL VS LOCAL TIME INDEX =====
At the time when data is saved to diagnostics, the global_idx == (local_idx-buffer) + K*chunk_size for some 
integer K. Meaning that the two indices are in sync, without any small offset. The very first chunk the local
index is merely offset by the buffer_size. 


===== SAVE_AT AND EXPORT_AT =====
All diagnostics needs an IntervalCounter as the member save_at
which dictates when data should be saved from the processor to the diagnostic

All diagnostics needs an IndexCounter as the member export_at
which dictates when data should be saved from the diagnostic to file. 

Both export_at and save_at will be indexes compared to the global time index. 

save_at should be/represent a sequence of non-overlapping intervals, where
each interval is (start, end), and the relevant signal should be saved
to the diagnostics between start:end (meaning end-exclusive).
To save state or instant diagnostics (where a continuous signal is not available)
the interval should be of length 1, so (start, start+1). 

When specifying save frequency, or giving just a list of indices (as you would for state or instant diagnostics)
it is interpreted as "save each X samples" or "save after X samples has passed". 
This means that with a save frequency of 200, the first save would be after globalIdx 199 has been processed; 
and as explained above, by the time data is saved to diagnostics, the globalIdx would then be 200.  


"""



class Logger:
    def __init__(self, sim_info):
        self.sim_info = sim_info
        self.diagnostics = {}

        self.upcoming_export = 0

    def __contains__(self, key):
        return key in self.diagnostics

    def __iter__(self):
        for diag_obj in self.diagnostics.values():
            yield diag_obj

    def items(self):
        for diag_name, diag_obj in self.diagnostics.items():
            yield diag_name, diag_obj

    def __getitem__(self, key):
        return self.diagnostics[key]

    def __setitem__(self, name, diagnostic):
        self.add_diagnostic(name, diagnostic)

    def prepare(self):
        self.update_next_export()
    
    def add_diagnostic(self, name, diagnostic):
        assert name not in self.diagnostics
        self.diagnostics[name] = diagnostic

    def save_data(self, processors, sig, idx, global_idx, last_block_on_chunk):
        for diag_name, diag in self.diagnostics.items():
            start, end = diag.next_save()
            num_samples = end - start

            if global_idx >= end:
                end_lcl = idx - (global_idx - end) - 1# or maybe start_lcl should be +1
                start_lcl = end_lcl - num_samples
                diag.save(processors, sig, (start_lcl, end_lcl), (start, end))
                diag.progress_save(end)
            elif last_block_on_chunk and global_idx > start:
                end_lcl = idx - 1
                start_lcl = idx - (global_idx-start) - 1
                diag.save(processors, sig, (start_lcl, end_lcl), (start, global_idx))
                diag.progress_save(global_idx)


    def update_next_export(self): 
        try:
            self.upcoming_export = np.amin([dg.next_export() for dg in self.diagnostics.values()])
        except ValueError:
            self.upcoming_export = np.inf

    def next_export(self):
        return self.upcoming_export
    
    def export_this_idx(self):
        diag_names = []
        for dg_name, dg in self.diagnostics.items():
            if dg.next_export() <= dg.save_at.saved_until: # all relevant data have been saved
                assert dg_name not in diag_names
                diag_names.append(dg_name)
        return diag_names

    def verify_same_export_settings(self, diag_dict):
        first_diag = diag_dict[list(diag_dict.keys())[0]]
        for dg in diag_dict.values():
            assert first_diag.export_function == dg.export_function
            assert first_diag.next_export() == dg.next_export()
            assert first_diag.keep_only_last_export == dg.keep_only_last_export
            assert first_diag.export_kwargs == dg.export_kwargs
            for prep1, prep2 in zip(first_diag.preprocess, dg.preprocess):
                assert len(prep1) == len(prep2)
                for pp_func1, pp_func2 in zip(prep1, prep2):
                    assert pp_func1.__name__ == pp_func2.__name__


    def export_single_diag(self, diag_name, fldr):
        diag_dict = {diag_name : self.diagnostics[diag_name]}
        #diag_dict = {proc.name : proc.diag[diag_name] for proc in processors if diag_name in proc.diag}
        self.verify_same_export_settings(diag_dict)

        one_diag_object = diag_dict[list(diag_dict.keys())[0]]
        exp_funcs = one_diag_object.export_function
        exp_kwargs = one_diag_object.export_kwargs
        preproc = one_diag_object.preprocess
        #export_time_idx = one_diag_object.next_save()[1]
        export_time_idx = one_diag_object.next_export()
        for exp_func, exp_kwarg, pp in zip(exp_funcs, exp_kwargs, preproc):
            exp_func(diag_name, diag_dict, export_time_idx, fldr, pp, print_method=self.sim_info.plot_output, **exp_kwarg)

        for diag in diag_dict.values():
            diag.progress_export()


    def dispatch(self, fldr):
        """
        processors is list of the processor objects
        time_idx is the global time index
        fldr is Path to figure folder
        """
        #if time_idx == self.next_export():
        if self.sim_info.plot_output != "none":
            while self.export_this_idx():
                for diag_name in self.export_this_idx():
                    self.export_single_diag(diag_name, fldr)
            self.update_next_export()


class IntervalCounter:
    def __init__(self, intervals, num_values = None):
        """
        intervals is an iterable or iterator where each entry is a tuple or list 
        of length 2, with start (inclusive) and end (exclusive) points of each interval
        np.ndarray of shape (num_intervals, 2) is also valid

        It is assumed that the intervals are strictly increasing, with no overlap. 
        meaning that ivs[i+1][0]>ivs[i][1] for all i. 
        """
        if isinstance(intervals, (list, tuple, np.ndarray)):
            if isinstance(intervals[0], (list, tuple, np.ndarray)):
                self.num_values = np.sum([iv[1]-iv[0] for iv in intervals])
            else: 
                self.num_values = len(intervals)
                intervals = [[idx-1, idx] for idx in intervals]
            self.intervals = iter(intervals)
            assert num_values is None
        else:
            self.intervals = intervals
            self.num_values = num_values

        self.saved_until = 0
        self.start, self.end = next(self.intervals, (np.inf, np.inf))
        

    @classmethod
    def from_frequency(cls, frequency, max_value, include_zero=False):
        num_values = int(np.ceil(max_value / frequency))

        start_value = 0
        if not include_zero:
            start_value += frequency
        return cls(zip(range(start_value-1, max_value, frequency), range(start_value, max_value+1, frequency)), num_values)


    # @classmethod
    # def from_frequency(cls, frequency, max_value, include_zero=True):
    #     num_values = int(np.ceil(max_value / frequency))

    #     start_value = 0
    #     if not include_zero:
    #         start_value += frequency
    #     return cls(zip(range(start_value, max_value, frequency), range(start_value+1, max_value+1, frequency)), num_values)

    def upcoming(self):
        return self.start, self.end

    def progress(self, progress_until):
        self.saved_until = progress_until
        self.start = progress_until
        assert self.end >= self.start
        if self.start == self.end:
            self.start, self.end = next(self.intervals, (np.inf, np.inf))




class IndexCounter():
    def __init__(self, idx_selection, tot_samples=None):
        """
        idx_selection is either an iterable of indices, or a single number
                        which is the interval between adjacent desired indices. 
        
        """
        try:
            self.orig_iterable = idx_selection
            self.idx_selection = iter(idx_selection)
        except TypeError:
            assert tot_samples is not None
            self.interval = idx_selection
            #self.idx_selection = it.count(idx_selection-1, idx_selection)
            self.idx_selection = iter(range(idx_selection, tot_samples+1, idx_selection))
            #self.idx_selection = it.count(idx_selection, idx_selection)
        self.upcoming_idx = next(self.idx_selection, np.inf)

    def progress(self):
        self.upcoming_idx = next(self.idx_selection, np.inf)

    def upcoming (self):
        return self.upcoming_idx

    def num_idx_until(self, until_value):
        raise ValueError
        #TODO check that this one works as expected
        try:
            return int(np.ceil(until_value / self.interval))
        except NameError:
            for i, idx in self.orig_iterable:
                if idx > until_value:
                    return i

        
class Diagnostic:
    export_functions = {}
    def __init__(
        self, 
        sim_info, 
        export_at,
        save_at,
        export_func,
        keep_only_last_export,
        export_kwargs,
        preprocess,
        ):
        """
        save_at_idx is an iterable which gives all indices for which to save data. 
                    Must be an integer multiple of the block size. (maybe change to 
                    must be equal or larger than the block size)
        """
        self.sim_info = sim_info

        assert isinstance(save_at, IntervalCounter)
        self.save_at = save_at

        if export_at is None:
            export_at = sim_info.export_frequency
        #if isinstance(export_at, (list, tuple, np.ndarray)):
        #    assert all([exp_at >= block_size for exp_at in export_at])
        #else:
            #assert export_at >= block_size
        
        self.export_at = IndexCounter(export_at, self.sim_info.tot_samples)

        if isinstance(export_func, str):
            export_func = [export_func]
        self.export_function = [type(self).export_functions[func_choice] for func_choice in export_func]

        if export_kwargs is None:
            export_kwargs = {}
        if isinstance(export_kwargs, dict):
            export_kwargs = [export_kwargs]
        assert len(export_kwargs) == len(self.export_function)
        self.export_kwargs = export_kwargs

        if preprocess is None:
            preprocess = [[] for _ in range(len(self.export_function))]
        elif callable(preprocess):
            preprocess = [preprocess]
        self.preprocess = [[pp] if callable(pp) else pp for pp in preprocess]
        assert len(self.preprocess) == len(self.export_function)

        self.keep_only_last_export = keep_only_last_export

        self.plot_data = {}

    def next_export(self):
        return self.export_at.upcoming()

    def next_save(self):
        return self.save_at.upcoming()

    def progress_save(self, progress_until):
        self.save_at.progress(progress_until)

    def progress_export(self):
        self.export_at.progress()

    @abstractmethod
    def save(self, processor, sig, chunk_interval, glob_interval):
        pass
        # self.get_property = op.attrgetter(property_name)
        # prop = self.get_property(processor)

        # signal = sig[signal_name]

    @abstractmethod
    def get_output(self):
        self.export_at.progress()

    def get_processed_output(self, time_idx, preprocess):
        """
        """
        output = self.get_output()
        for pp in preprocess:
            output = pp(output)
        return output


class SignalDiagnostic(Diagnostic):
    export_functions = {
        "plot" : dplot.function_of_time_plot,
        "npz" : dplot.savenpz,
        "text" : dplot.txt,
        "wav" : dplot.create_audio_files,
        "spectrum" : dplot.spectrum_plot, 
    }
    def __init__ (
        self,
        sim_info,
        export_at=None,
        save_at = None, 
        export_func = "plot",
        keep_only_last_export = None,
        export_kwargs = None,
        preprocess = None,
    ):
        if save_at is None:
            save_at = IntervalCounter(((0,sim_info.tot_samples),))
        if keep_only_last_export is None:
            keep_only_last_export = False
        super().__init__(sim_info, export_at, save_at, export_func, keep_only_last_export, export_kwargs, preprocess)

        self.plot_data["xlabel"] = "Samples"
        self.plot_data["ylabel"] = ""
        self.plot_data["title"] = ""

    def get_processed_output(self, time_idx, preprocess):
        output, time_indices = get_values_up_to_idx(self.get_output(), time_idx)
        for pp in preprocess:
            output = pp(output)
        return output, time_indices


class StateDiagnostic(Diagnostic):
    export_functions = {
        "plot" : dplot.function_of_time_plot,
        "npz" : dplot.savenpz,
    }
    def __init__ (
        self,
        sim_info,
        export_at=None,
        save_frequency=None,
        export_func = "plot",
        keep_only_last_export=False,
        export_kwargs = None,
        preprocess = None,
    ):
        if save_frequency is None:
            raise NotImplementedError
            #save_frequency = block_size
        super().__init__(sim_info, export_at, 
                        IntervalCounter.from_frequency(save_frequency, sim_info.tot_samples,include_zero=False),
                        export_func, 
                        keep_only_last_export,
                        export_kwargs,
                        preprocess)
        self.plot_data["xlabel"] = "Samples"
        self.plot_data["ylabel"] = ""
        self.plot_data["title"] = ""

    def get_processed_output(self, time_idx, preprocess):
        output, time_indices = get_values_from_selection(self.get_output(), self.time_indices, time_idx)
        for pp in preprocess:
            output = pp(output)
        return output, time_indices


class InstantDiagnostic(Diagnostic):
    export_functions = {
        "plot" : dplot.plot_ir,
        "matshow" : dplot.matshow, 
        "npz" : dplot.savenpz,
        "text" : dplot.txt, 
    }
    def __init__ (
        self,
        sim_info,
        save_at = None, 
        export_func = "plot",
        keep_only_last_export = False,
        export_kwargs = None,
        preprocess = None,
    ):
        if save_at is None:
            save_freq = sim_info.export_frequency
            save_at = IntervalCounter.from_frequency(save_freq, sim_info.tot_samples, include_zero=False)
            export_at = save_freq
            #export_at = IndexCounter(save_freq)
        else:
            # Only a single list, or a scalar. save_at can't be intervals
            export_at = save_at
            if isinstance(save_at, (tuple, list, np.ndarray)):
                assert not isinstance(save_at[0], (tuple, list, np.ndarray))
                save_at = IntervalCounter(save_at)
            else:
                save_at = IntervalCounter.from_frequency(save_at, sim_info.tot_samples, include_zero=False)
        super().__init__(sim_info, export_at, save_at, export_func, keep_only_last_export, export_kwargs, preprocess)



# def attritemgetter(name):
#     assert name[0] != "["
#     attributes = name.replace("]", "").replace("'", "")
#     attributes = attributes.split(".")
#     attributes = [attr.split("[") for attr in attributes]

#     def getter (obj):
#         for sub_list in attributes:
#             obj = getattr(obj, sub_list[0])
#             for item in sub_list[1:]:
#                 obj = obj[item]
#         return obj
#     return getter

def attritemgetter(name):
    """
    If you have a dictionary with strings, use the name "dict_obj['key']"
    Without apostrophes, the key is assumed to be a list index, and is converted to integer.

    TODO: Possibly allow for objects other than integers, such as indexing arrays
    
    """
    assert name[0] != "["
    attributes = name.replace("]", "")
    attributes = attributes.split(".")
    attributes = [attr.split("[") for attr in attributes]

    def getter (obj):
        for sub_list in attributes:
            obj = getattr(obj, sub_list[0])
            for item in sub_list[1:]:
                if item[0] == "'":
                    if not item[-1] == "'":
                        raise ValueError("Unmatched apostrophes")
                    item = item.replace("'", "")
                else:
                    item = int(item)
                obj = obj[item]
        return obj
    return getter




def get_values_up_to_idx(signal, max_idx):
    """
    gives back signal values that correspond to time_values less than max_idx, 
    and signal values that are not nan

    max_idx is exlusive
    """
    signal = np.atleast_2d(signal)
    assert signal.ndim == 2

    time_indices = np.arange(max_idx)
    signal = signal[:,:max_idx]

    nan_filter = np.logical_not(np.isnan(signal))
    assert np.isclose(nan_filter, nan_filter[0, :]).all()
    nan_filter = nan_filter[0, :]

    time_indices = time_indices[nan_filter]
    signal = signal[:,nan_filter]
    return signal, time_indices

def get_values_from_selection(signal, time_indices, max_idx):
    """
    gives back signal values that correspond to time_values less than max_idx, 
    and signal values that are not nan

    max_idx is exlusive
    """
    signal = np.atleast_2d(signal)
    assert signal.ndim == 2
    nan_filter = np.logical_not(np.isnan(signal))
    assert np.isclose(nan_filter, nan_filter[0, :]).all()
    nan_filter = nan_filter[0, :]

    assert time_indices.shape[-1] == signal.shape[-1]
    time_indices = time_indices[nan_filter]
    signal = signal[:,nan_filter]

    if len(time_indices) == 0:
        assert signal.shape[-1] == 0
        return signal, time_indices

    above_max_idx = np.argmax(time_indices >= max_idx)
    if above_max_idx == 0:
        if np.logical_not(above_max_idx).all():
            above_max_idx = len(time_indices)

    time_indices = time_indices[:above_max_idx]
    signal = signal[:,:above_max_idx]
    return signal, time_indices



import pytest
import hypothesis as hyp
import hypothesis.strategies as st

import numpy as np
import scipy.signal.windows as win

import aspsim.signal.sources as sources

# @pytest.mark.parametrize("src", [sources.SineSource(1,100,8000)])
# def test_source_blocks(src, ):


def test_sine_blocks():
    numSamples = 1000
    numBlocks = 50
    block_size = int(numSamples / numBlocks)
    freq = 100
    sr = 8000

    np.random.seed(0)
    src = sources.SineSource(1, 1, freq, sr)
    ref = src.get_samples(numSamples)

    np.random.seed(0)
    source = sources.SineSource(1, 1, freq, sr)
    out = np.zeros(numSamples)
    for i in range(0, numSamples, block_size):
        out[i : i + block_size] = source.get_samples(block_size)

    assert np.mean(out - ref) < 1e-6


def test_noisesrc_blocks():
    N = 10000
    numBlocks = 100
    block_size = int(N / numBlocks)
    freqs = (100, 400)
    sr = 8000

    np.random.seed(0)
    source = sources.BandlimitedNoiseSource(1, 1, freqs, sr)
    sig1 = source.get_samples(N)
    # sig1 = util.pow2db(np.abs(np.fft.rfft(sig1))**2)

    np.random.seed(0)
    source2 = sources.BandlimitedNoiseSource(1, 1, freqs, sr)
    sig2 = np.zeros((1, N))
    for i in range(N // 100):
        sig2[:, i * block_size : (i + 1) * block_size] = source2.get_samples(block_size)

    assert np.sum(np.abs(sig1 - sig2)) < 1e-6


def test_multisine_blocks():
    numSamples = 10000
    numBlocks = 100
    block_size = int(numSamples / numBlocks)
    freqs = np.linspace(100, 1000, 10)
    sr = 8000

    rng = np.random.default_rng(1)
    source1 = sources.MultiSineSource(1, 1, freqs, sr, rng=rng)
    sig1 = source1.get_samples(numSamples)

    rng = np.random.default_rng(1)
    source2 = sources.MultiSineSource(1, 1, freqs, sr, rng=rng)
    sig2 = np.zeros((1, numSamples))
    for i in range(numBlocks):
        sig2[:, i * block_size : (i + 1) * block_size] = source2.get_samples(block_size)

    assert np.sum(np.abs(sig1 - sig2)) < 1e-6



@hyp.given(num_samples = st.integers(min_value=64,max_value=1024),
            num_channels = st.integers(min_value=1, max_value=5),
            block_size = st.integers(min_value=1, max_value = 64),
            seq_len = st.integers(min_value = 1, max_value=1500))
def test_sequence_source_blocks(num_samples, num_channels, block_size, seq_len):
    num_samples -= num_samples % block_size
    num_blocks = num_samples // block_size

    rng = np.random.default_rng(1)
    seq = rng.normal(0, 1, (num_channels,seq_len))

    src1 = sources.Sequence(seq, end_mode="repeat")
    sig1 = src1.get_samples(num_samples)

    src2 = sources.Sequence(seq, end_mode="repeat")
    sig2 = np.zeros((num_channels, num_samples))

    for i in range(num_blocks):
        sig2[:, i * block_size : (i + 1) * block_size] = src2.get_samples(block_size)
    assert sig1.shape == (num_channels, num_samples)
    assert sig2.shape == (num_channels, num_samples)
    assert np.mean(np.abs(sig1 - sig2)) < 1e-10

@hyp.given(num_channels = st.integers(min_value=1, max_value=5),
            num_repeats = st.integers(min_value=2, max_value = 8),
            seq_len = st.integers(min_value = 1, max_value=1500))
def test_sequence_source_repeat(num_channels, num_repeats, seq_len):
    rng = np.random.default_rng(1)
    seq = rng.normal(0, 1, (num_channels,seq_len))

    src = sources.Sequence(seq, end_mode="repeat")
    sig = src.get_samples(seq_len*num_repeats)

    for r in range(num_repeats):
        assert np.allclose(seq, sig[:, r*seq_len:(r+1)*seq_len])



def test_gold_sequence_remembering_state():
    src = sources.GoldSequenceSource(1,1, 11)
    N = 100
    sig1 = src.get_samples(N)
    sig2 = src.get_samples(N)
    sig = np.concatenate((sig1, sig2), axis=-1)

    src = sources.GoldSequenceSource(1,1,11)
    sigCompare = src.get_samples(2 * N)

    assert np.allclose(sigCompare, sig)





#@hyp.settings(deadline=None)
#@hyp.given(bs = st.integers(min_value=1, max_value=5))
@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5),
            num_samples = st.integers(min_value=10, max_value=100),
            num_channels = st.integers(min_value=1, max_value=3))

def test_pulse_train(bs, num_samples, num_channels):
    #import matplotlib.pyplot as plt
    rng = np.random.default_rng()
    period = rng.integers(low=1, high=10, size=num_channels)
    dly = rng.integers(low=0, high=30, size=num_channels)
    amp = rng.standard_normal(size=num_channels)

    pulse_src = sources.PulseTrain(num_channels, amp, period, dly)
    sig = np.zeros((num_channels,num_samples))
    for i in range(num_samples // bs):
        sig[:,i*bs:(i+1)*bs] = pulse_src.get_samples(bs)
    sig[:,(i+1)*bs:] = pulse_src.get_samples(num_samples-(i+1)*bs)

    for j in range(num_channels):
        for i in range(num_samples):
            if (i - dly[j]) % period[j] == 0:
                assert sig[j,i] == amp[j]
            else:
                assert sig[j,i] == 0
    #plt.plot(sig.T)
    #plt.show()





# def test_autocorr_of_AutocorrSource():
#     #autocorr = np.array([[2,0.5,0.5,0.2,0.1]], dtype=float)
#     autocorr = np.zeros((1, 25))
#     autocorr[0,0] = 1
#     autocorr[0,4] = 0.5
#     src = sources.AutocorrSource(1, autocorr)
#     sig = src.get_samples(int(1e6))
#     ac = cr.Autocorrelation(1, 5*autocorr.shape[-1], autocorr.shape[0])
#     ac.update(sig)
#     print(ac.corr.state)
#     pass # todo: calculate the power of the white noise. The relative sizes of the resulting autocorr is correct



# def test_bandlim_noise():
#     # import matplotlib.pyplot as plt
#     sr = 8000
#     freqLim = (200, 250)
#     src = sources.BandlimitedNoiseSource(1, freqLim, sr)
#     signal = src.get_samples(10000)
#     f, p = spsig.welch(signal, sr, nperseg=2048, window="hamming")
    # p = np.abs(np.fft.fft(np.squeeze(signal)))**2
    # plt.plot(10*np.log10(p))
    # plt.plot(f, 10*np.log10(np.squeeze(p)))
    # plt.show()


# def test_gold_sequence_autocorrelation():
#     import matplotlib.pyplot as plt

#     src = sources.GoldSequenceSource(11)
#     N = 10000
#     sig = np.squeeze(src.get_samples(N))
#     corr = np.correlate(sig, sig, mode="full")
#     plt.plot(corr)
    # plt.show()


# def test_MLS_autocorrelation():
#     import matplotlib.pyplot as plt

#     src = sources.MLS(11, [11, 2, 0])
#     N = 1000
#     sig = np.squeeze(src.get_samples(N))
#     corr = np.correlate(sig, sig, mode="full")
#     plt.plot(corr)
    # plt.show()


# def test_preferred_pair_autocorrelation():
#     import matplotlib.pyplot as plt

#     src = sources.MLS(11, [11, 2, 0])
#     src2 = sources.MLS(11, [11, 8, 5, 2, 0])
#     N = 1000
#     sig = np.squeeze(src.get_samples(N))
#     sig2 = np.squeeze(src2.get_samples(N))
#     corr = np.correlate(sig, sig2, mode="full")
#     plt.plot(corr)
    # plt.show()


# def test_gold_autocorrelation_2():
#     import matplotlib.pyplot as plt

#     src = sources.MLS(8, [8, 6, 5, 3, 0], state=[0, 0, 0, 0, 0, 0, 0, 1])
#     src2 = sources.MLS(8, [8, 6, 5, 2, 0], state=[0, 0, 0, 0, 0, 0, 0, 1])
#     N = 256
#     sig = np.squeeze(src.get_samples(N))
#     sig2 = np.squeeze(src2.get_samples(N))
#     corr = np.correlate(sig, sig2, mode="full")
#     plt.plot(corr)
#     # plt.show()



@pytest.fixture
def chirpSetup():
    samplesToSweep = 3999
    minFreq = 100
    maxFreq = 3000
    sr = 8000
    amp = 1
    src = sources.LinearChirpSource(amp, (minFreq, maxFreq), samplesToSweep, sr)
    return src, sr, minFreq, maxFreq, samplesToSweep


def test_chirp_blocks():
    numSamples = 10000
    numBlocks = 100
    block_size = int(numSamples / numBlocks)
    freqs = (100, 400)
    numToSweep = 1000
    sr = 8000

    np.random.seed(0)
    source = sources.LinearChirpSource(1, 1, freqs, numToSweep, sr)
    noise = source.get_samples(numSamples)

    np.random.seed(0)
    source2 = sources.LinearChirpSource(1, 1, freqs, numToSweep, sr)
    noise2 = np.zeros(numSamples)
    for i in range(numBlocks):
        noise2[i * block_size : (i + 1) * block_size] = source2.get_samples(block_size)

    assert np.sum(np.abs(noise - noise2)) < 1e-6

def test_chirp_min_frequency_property(chirpSetup):
    src, sr, minFreq, maxFreq, samplesToSweep = chirpSetup
    signal = src.get_samples(samplesToSweep * 10)
    assert src.freq - minFreq < 1e-6


def test_chirp_max_frequency_property(chirpSetup):
    src, sr, minFreq, maxFreq, samplesToSweep = chirpSetup
    signal = src.get_samples(samplesToSweep * 9)
    assert src.freq - maxFreq < 1e-6


def test_chirp_min_frequency_fft(chirpSetup):
    src, sr, minFreq, maxFreq, samplesToSweep = chirpSetup
    block_size = 128
    fftsize = 2 ** 14
    src.get_samples(int(samplesToSweep * 20 - block_size / 2))
    signal = src.get_samples(block_size)
    freq = get_freq_of_sine(signal, sr, block_size, fftsize)

    freqBinDif = sr / fftsize
    blockFreqChange = (block_size / (samplesToSweep * 2)) * (maxFreq - minFreq)
    assert freq - minFreq < blockFreqChange


def test_chirp_max_frequency_fft(chirpSetup):
    src, sr, minFreq, maxFreq, samplesToSweep = chirpSetup
    block_size = 128
    fftsize = 2 ** 14
    src.get_samples(int(samplesToSweep * 19 - block_size / 2))
    signal = src.get_samples(block_size)
    freq = get_freq_of_sine(signal, sr, block_size, fftsize)

    freqBinDif = sr / fftsize
    blockFreqChange = (block_size / (samplesToSweep * 2)) * (maxFreq - minFreq)
    assert freq - maxFreq < blockFreqChange


def get_freq_of_sine(signal, sr, block_size, fftsize):
    signal *= win.hamming(block_size)
    freqs = np.fft.rfft(signal, n=fftsize)
    maxBin = np.argmax(freqs)
    freq = sr * maxBin / (2 * fftsize)
    return freq



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import pyroomacoustics as pra

import aspcore.fouriertransform as ft

import aspsim.room.region as region
import aspsim.room.directional as directional
from aspsim.simulator import SimulatorSetup
import aspsim.room.roomimpulseresponse as rir
import aspsim.room.generatepoints as gp


import aspsim.utilities as utils

def _setup_ism(fig_folder, samplerate):
    setup = SimulatorSetup(fig_folder)
    setup.sim_info.samplerate = samplerate
    setup.sim_info.tot_samples =  samplerate
    setup.sim_info.sim_chunk_size = 2*samplerate
    setup.sim_info.sim_buffer = samplerate
    setup.sim_info.export_frequency =  samplerate
    setup.sim_info.save_source_contributions = False
    setup.sim_info.randomized_ism = False

    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [4, 3, 3]
    setup.sim_info.room_center = [-1, 0, 0]
    setup.sim_info.rt60 = 0.25
    setup.sim_info.max_room_ir_length = samplerate // 2
    return setup


def test_directional_microphones_are_correct_with_regards_to_reciprocity():
    """
    When the directional microphone is pointed towards the source or one of the two reflections, the impulse response for that particular reflection should be identical to the impulse response of the omni-directional microphone.
    
    change src_azimuth in the calculation of mic_dir to ref1_azimuth or ref2_azimuth to test the reflections. 
    """
    pos_src = np.array([[1, 1, 0]])
    pos_mic = np.array([[-2.5, -3, 0]])
    room_size = np.array([150, 8, 150])
    room_center = np.array([0, 0, 0])
    ir_len = 1024
    sr = 4000
    e_absorbtion = 0.5
    max_order = 1
    min_dly = 0

    src_angle = pos_src - pos_mic
    src_angle = src_angle / np.linalg.norm(src_angle, axis=-1, keepdims=True)
    src_azimuth = np.arctan2(src_angle[0,1], src_angle[0,0])


    wall_bottom = room_center[1] - room_size[1] / 2
    wall_top = room_center[1] + room_size[1] / 2
    y1 = pos_mic[0,1] - wall_bottom
    y2 = pos_src[0,1] - wall_bottom
    ref1_azimuth = -np.arctan2((y1 + y2), (np.abs(pos_src[0,0] - pos_mic[0,0]))) #correct for bottom walls

    y1 = wall_top - pos_mic[0,1]
    y2 = wall_top - pos_src[0,1]
    ref2_azimuth = np.arctan2((y1 + y2), (np.abs(pos_src[0,0] - pos_mic[0,0]))) #correct for top walls

    mic_dir = gp.spherical2cart(np.ones(1,), np.array([[src_azimuth, np.pi/2]]))

    ir_omni = rir.ir_room_image_source_3d(pos_src,
    pos_mic,
    room_size,
    room_center,
    ir_len,
    sr,
    e_absorbtion,
    max_order,
    min_dly,
    dir_type_mic = None,
    dir_dir_mic = None,
    randomized_ism = False,
    calculate_metadata=False,
    verbose=False)

    ir_dir = rir.ir_room_image_source_3d(pos_src,
    pos_mic,
    room_size,
    room_center,
    ir_len,
    sr,
    e_absorbtion,
    max_order,
    min_dly,
    dir_type_mic = ["cardioid"],
    dir_dir_mic = mic_dir,
    randomized_ism = False,
    calculate_metadata=False,
    verbose=False)


    fig, ax = plt.subplots(1, 1)
    corner = [c - sz/2 for c, sz in zip(room_center[:2], room_size[:2])]
    width = room_size[0]
    height = room_size[1]
    ax.add_patch(patches.Rectangle(corner, width, height, edgecolor="k", facecolor="none", linewidth=2, alpha=1))
    
    ax.plot(pos_mic[:,0], pos_mic[:,1], "x", label="Mic")
    ax.plot(pos_src[:,0], pos_src[:,1], "o", label="Src")
    ax.plot([pos_mic[0,0], pos_mic[0,0]+3*mic_dir[0,0]], [pos_mic[0,1], pos_mic[0,1]+3*mic_dir[0,1]], label="Mic dir", linestyle="--")  

    ax.axis("equal")
    margin = 0.5
    ax.set_xlim([np.min((pos_mic[:,0], pos_src[:,0])) -margin, np.max((pos_mic[:,0], pos_src[:,0])) -margin])
    
    fig, ax = plt.subplots(1, 1)
    ax.plot(ir_omni[0,0,:], label="Omni")
    ax.plot(ir_dir[0,0,:], label="Dir")
    ax.legend()
    plt.show()


def test_cardioid_microphone_facing_plane_wave_gives_same_response_as_omni_microphone():
    rng = np.random.default_rng(10)
    sr = 1000

    pos_mic = np.zeros((1,3))

    pw_dir = np.array([[1, 1, 1]])
    pw_dir = pw_dir / np.linalg.norm(pw_dir) # incoming from that direction, not propagation direction
    mic_dir = pw_dir
    
    src_distance = 500
    pos_src = pw_dir * src_distance

    setup = _get_default_simulator_setup(sr)
    prop_delay = np.linalg.norm(pos_src) / setup.sim_info.c
    rir_len = int(2 * prop_delay * sr)

    setup.sim_info.room_size = [3*src_distance, 3*src_distance, 3*src_distance]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.rt60 =  0
    setup.sim_info.max_room_ir_length = rir_len
    setup.sim_info.sim_buffer = rir_len
    setup.sim_info.extra_delay = 100

    setup.add_mics("omni", pos_mic)
    setup.add_mics("cardioid", pos_mic, directivity_type=["cardioid"], directivity_dir=mic_dir)
    setup.add_controllable_source("src", pos_src)
    sim = setup.create_simulator()

    fig, axes = plt.subplots(2,1)
    axes[0].plot(sim.arrays.paths["src"]["omni"][0,0,:])
    axes[0].plot(sim.arrays.paths["src"]["cardioid"][0,0,:])
    axes[1].plot(sim.arrays.paths["src"]["cardioid"][0,0,:] - sim.arrays.paths["src"]["omni"][0,0,:])
    plt.show()

    mse = np.mean(np.abs(sim.arrays.paths["src"]["omni"][0,0,:] - sim.arrays.paths["src"]["cardioid"][0,0,:])**2) / \
        np.mean(np.abs(sim.arrays.paths["src"]["omni"][0,0,:])**2)
    
    assert mse < 1e-10

    #num_freqs = setup.sim_info.max_room_ir_length
    #fpaths, freqs = sim.arrays.get_freq_paths(num_freqs, sr)

    #return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], sim.arrays.paths["src"]["omni"], \
    #        sim.arrays["cardioid"].pos, fpaths["src"]["cardioid"][...,0], freqs, sim.sim_info


def test_cardioid_microphone_response_compared_to_omni_for_plane_wave_pyroomacoustics():
    samplerate = 1000

    #Plane wave incoming from the following direction, which is negative of propagation direction
    pw_azimuth = 0.234 * np.pi # arbitrary angle
    pw_colatitude = np.pi / 2 + 0.4 * np.pi
    pw_dir = np.array([np.cos(pw_azimuth) * np.sin(pw_colatitude), np.sin(pw_azimuth) * np.sin(pw_colatitude), np.cos(pw_colatitude)])
    
    src_distance = 500
    room_size = np.array([3*src_distance, 3*src_distance, 3*src_distance]) # make room big enough
    room_center = room_size / 2
    pos_mic = room_center[:,np.newaxis]
    pos_src = room_center + pw_dir * src_distance # Put source very far away to approximate plane wave

    # Direct microphone towards the plane wave
    dir_obj = pra.directivities.CardioidFamily(
        orientation=pra.directivities.DirectionVector(azimuth=pw_azimuth, colatitude=pw_colatitude, degrees=False),
        pattern_enum=pra.directivities.DirectivityPattern.CARDIOID,
    )

    # RIR for cardioid microphone
    room = pra.ShoeBox(room_size, fs=samplerate, max_order=0, use_rand_ism = False)
    room.add_source(pos_src)
    mics = pra.MicrophoneArray(pos_mic, samplerate, directivity=dir_obj)
    room.add_microphone_array(mics)

    room.compute_rir()
    rir_cardioid = room.rir[0][0]

    # RIR for omni microphone
    room = pra.ShoeBox(room_size, fs=samplerate, max_order=0, use_rand_ism = False)
    room.add_source(pos_src)
    mics = pra.MicrophoneArray(pos_mic, samplerate)
    room.add_microphone_array(mics)
    room.compute_rir()
    rir_omni = room.rir[0][0]

    fig, axes= plt.subplots(2, 1)
    axes[0].plot(rir_cardioid, label = "Cardioid")
    axes[0].plot(rir_omni, label = "Omni")
    axes[0].legend()
    axes[1].plot(rir_cardioid - rir_omni, label = "Difference")
    axes[1].legend()
    plt.show()



def test_cardioid_microphone_response_to_plane_wave_of_different_heights_pyroomacoustics():
    samplerate = 1000

    #Plane wave incoming from the following direction, which is negative of propagation direction
    pw_azimuth = 0.234 * np.pi # arbitrary angle
    pw_colatitude = np.pi / 2
    pw_dir = np.array([np.cos(pw_azimuth) * np.sin(pw_colatitude), np.sin(pw_azimuth) * np.sin(pw_colatitude), np.cos(pw_colatitude)])
    
    src_distance = 500
    room_size = np.array([3*src_distance, 3*src_distance, 3*src_distance]) # make room big enough
    room_center = room_size / 2
    pos_mic = room_center[:,np.newaxis]
    pos_src = room_center + pw_dir * src_distance # Put source very far away to approximate plane wave

    

    # RIR for cardioid microphone
    # Direct microphone towards the plane wave
    dir_obj = pra.directivities.CardioidFamily(
        orientation=pra.directivities.DirectionVector(azimuth=pw_azimuth, colatitude=pw_colatitude, degrees=False), pattern_enum=pra.directivities.DirectivityPattern.CARDIOID,
    )
    room = pra.ShoeBox(room_size, fs=samplerate, max_order=0, use_rand_ism = False)
    room.add_source(pos_src)
    mics = pra.MicrophoneArray(pos_mic, samplerate, directivity=dir_obj)
    room.add_microphone_array(mics)

    room.compute_rir()
    rir_cardioid = room.rir[0][0]

    # RIR for plane wave with slightly rotated mic
    # Direct microphone not exactly towards the plane wave
    dir_obj = pra.directivities.CardioidFamily(
        orientation=pra.directivities.DirectionVector(azimuth=pw_azimuth, colatitude=pw_colatitude + 0.3*np.pi, degrees=False), pattern_enum=pra.directivities.DirectivityPattern.CARDIOID,
    )
    room = pra.ShoeBox(room_size, fs=samplerate, max_order=0, use_rand_ism = False)
    room.add_source(pos_src)
    mics = pra.MicrophoneArray(pos_mic, samplerate, directivity=dir_obj)
    room.add_microphone_array(mics)
    room.compute_rir()
    rir_omni = room.rir[0][0]

    fig, axes= plt.subplots(2, 1)
    axes[0].plot(rir_cardioid, label = "Cardioid")
    axes[0].plot(rir_omni, label = "Omni")
    axes[0].legend()
    axes[1].plot(rir_cardioid - rir_omni, label = "Difference")
    axes[1].legend()
    plt.show()



def _generate_plane_wave_with_ism(sr, pw_dir, pos_mic=None, pos_dir=None, mic_dir=None):
    rng = np.random.default_rng(10)
    side_len = 0.5
    num_mic = 40

    assert np.allclose(np.linalg.norm(pw_dir), 1)

    #pos_mic = np.zeros((num_mic, 3))
    if pos_mic is None:
        pos_mic = rng.uniform(low=-side_len/2, high=side_len/2, size=(num_mic, 3))

    if pos_dir is None:
        pos_dir = np.zeros((1,3))
        mic_dir = np.array([[1,0,0]])
    
    pos_src = pw_dir * 500 #np.array([[0,200,0]])

    setup = _get_default_simulator_setup(sr)
    prop_delay = np.linalg.norm(pos_src) / setup.sim_info.c
    rir_len = int(2 * prop_delay * sr)

    setup.sim_info.room_size = [2000, 2000, 2000]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.rt60 =  0
    setup.sim_info.max_room_ir_length = rir_len
    setup.sim_info.sim_buffer = rir_len
    setup.sim_info.extra_delay = 100

    setup.add_mics("omni", pos_mic)
    setup.add_mics("cardioid", pos_dir, directivity_type=pos_dir.shape[0]*["cardioid"], directivity_dir=mic_dir)
    setup.add_controllable_source("src", pos_src)
    sim = setup.create_simulator()

    num_freqs = setup.sim_info.max_room_ir_length
    fpaths, freqs = sim.arrays.get_freq_paths(num_freqs, sr)

    return sim.arrays["omni"].pos, fpaths["src"]["omni"][...,0], sim.arrays.paths["src"]["omni"], \
            sim.arrays["cardioid"].pos, fpaths["src"]["cardioid"][...,0], freqs, sim.sim_info





def _get_default_simulator_setup(sr):
    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 =  0.25
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 60
    setup.sim_info.plot_output = "none"
    return setup




def test_differential_microphone_filter_gives_same_result_as_pyroomacoustics_for_approximate_plane_wave_sound_field():
    sr = 1000
    num_rotations = 30

    pos_mic = np.zeros((1,3))
    pos_mic_tiled = np.tile(pos_mic, (num_rotations,1))
    mic_angle = np.linspace(0, 2*np.pi, num_rotations, endpoint=False)
    mic_direction = utils.spherical2cart(np.ones(num_rotations), np.stack((mic_angle, np.pi/2 * np.ones(num_rotations)),axis=-1))

    diff_mic_distance = 0.01
    pos_mic_secondary = pos_mic - diff_mic_distance * mic_direction

    pos_src = np.array([[0,90,0]])

    setup = SimulatorSetup()
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [200, 200, 200]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.max_room_ir_length = 2*sr
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = sr // 2
    setup.sim_info.plot_output = "none"
    setup.sim_info.rt60 = 0

    setup.add_mics("omni", pos_mic)
    setup.add_mics("secondary_mics", pos_mic_secondary)
    dir_type = num_rotations*["cardioid"]
    setup.add_mics("cardioid", pos_mic_tiled, directivity_type=dir_type, directivity_dir=mic_direction)
    setup.add_controllable_source("src", pos_src)

    sim = setup.create_simulator()

    signal_main = np.tile(sim.arrays.paths["src"]["omni"][0,:,:], (num_rotations,1))
    signal_secondary = sim.arrays.paths["src"]["secondary_mics"][0,:,:]
    filter_below = 200
    cardioid_signal = directional.differential_cardioid_microphone(signal_main, signal_secondary, diff_mic_distance, 500, sim.sim_info.c, sim.sim_info.samplerate, filter_below=filter_below)
    
    power_per_angle_pra = np.mean(np.abs(sim.arrays.paths["src"]["cardioid"][0,:,:])**2, axis=-1)
    power_per_angle_diff = np.mean(np.abs(cardioid_signal)**2, axis=-1)
    power_error = np.mean(np.abs(power_per_angle_pra - power_per_angle_diff))

    fig, ax = plt.subplots(1,1)
    ax.set_title(f"Mean power error: {power_error}")
    ax.set_ylabel("Response power")
    ax.set_xlabel("Microphone angle")
    ax.plot(mic_angle, power_per_angle_pra, label="Pyroomacoustics")
    ax.plot(mic_angle, power_per_angle_diff, label="Differential")
    ax.legend()

    fig, ax = plt.subplots(1,1)
    example_idx = 5
    mse_time_response = np.mean(np.abs(cardioid_signal - sim.arrays.paths["src"]["cardioid"][0,:,:])**2) / np.mean(np.abs(sim.arrays.paths["src"]["cardioid"][0,:,:])**2)
    ax.set_title(f"Mean square error: {mse_time_response}")
    ax.set_xlabel("Time (samples)")
    ax.set_ylabel("Amplitude")
    ax.plot(sim.arrays.paths["src"]["cardioid"][0,example_idx,:], label="Pyroomacoustics")
    ax.plot(cardioid_signal[example_idx,:], label="Differential")
    ax.plot(sim.arrays.paths["src"]["omni"][0,0,:], label="Pra omni")
    ax.legend()

    fig, axes = plt.subplots(3,1, sharex=True, figsize=(6,8))

    axes[2].set_xlabel("Frequency")
    axes[0].set_ylabel("Absolute")
    axes[1].set_ylabel("Real")
    axes[2].set_ylabel("Imag")
    freqs = ft.get_real_freqs(cardioid_signal.shape[-1], sim.sim_info.samplerate)
    freq_response_pra = ft.rfft(sim.arrays.paths["src"]["cardioid"][0,example_idx,:])
    freq_response_diff = ft.rfft(cardioid_signal[example_idx,:])
    freq_mask = freqs > filter_below
    freq_mask[-5:] = False
    mse = np.mean(np.abs(freq_response_pra[freq_mask] - freq_response_diff[freq_mask])**2) / np.mean(np.abs(freq_response_pra[freq_mask])**2)
    axes[0].set_title(f"Mean square error: {mse}")

    axes[0].plot(freqs, np.abs(freq_response_pra), label="Pyroomacoustics")
    axes[0].plot(freqs, np.abs(freq_response_diff), label="Differential")
    axes[0].plot(freqs, np.abs(ft.rfft(sim.arrays.paths["src"]["omni"][0,0,:])), label="Pra omni")

    axes[1].plot(freqs, np.real(freq_response_pra), label="Pyroomacoustics")
    axes[1].plot(freqs, np.real(freq_response_diff), label="Differential")
    #axes[1].plot(freqs, np.real(ft.rfft(sim.arrays.paths["src"]["omni"][0,0,:])), label="Pra omni")

    axes[2].plot(freqs, np.imag(freq_response_pra), label="Pyroomacoustics")
    axes[2].plot(freqs, np.imag(freq_response_diff), label="Differential")
    #axes[2].plot(freqs, np.imag(ft.rfft(sim.arrays.paths["src"]["omni"][0,0,:])), label="Pra omni")

    plt.show()


import pytest
import numpy as np
import aspsim.utilities as util
from hypothesis import given
import hypothesis.strategies as st


@given(
    st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100)
)
def test_calc_block_sizes_totalEqualToNumSamples(numSamples, block_size):
    startIdx = np.random.randint(0, block_size)
    sizes = util.calc_block_sizes(numSamples, startIdx, block_size)
    assert np.sum(sizes) == numSamples


@given(
    st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100)
)
def test_calc_block_sizes_maxValueEqualToBlockLength(numSamples, block_size):
    startIdx = np.random.randint(0, block_size)
    sizes = util.calc_block_sizes(numSamples, startIdx, block_size)
    assert np.max(sizes) <= block_size


@given(
    st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100)
)
def test_calc_block_sizes_noZeroValues(numSamples, block_size):
    startIdx = np.random.randint(0, block_size)
    sizes = util.calc_block_sizes(numSamples, startIdx, block_size)
    assert np.min(sizes) > 0


@given(
    st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100)
)
def test_calc_block_sizes_allMiddleValuesEqualToblock_size(numSamples, block_size):
    startIdx = np.random.randint(0, block_size)
    sizes = util.calc_block_sizes(numSamples, startIdx, block_size)
    print(sizes)
    if len(sizes) >= 3:
        assert np.allclose(sizes[1:-1], block_size)


@given(
    st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100)
)
def test_calc_block_sizes_firstValueCorrect(numSamples, block_size):
    startIdx = np.random.randint(0, block_size)
    sizes = util.calc_block_sizes(numSamples, startIdx, block_size)
    assert sizes[0] == np.min((block_size - startIdx, numSamples))



import numpy as np
import scipy.spatial.distance as distfuncs
import scipy.signal as spsig
import pyroomacoustics as pra
import pyroomacoustics.directivities as pradir

import aspsim.room.generatepoints as gp


# def setup_ir(arrays, sim_info):
#     """Generates the impulse responses between all sources and microphones

#     The method uses the path_type attribute to determine the type of propagation
    
#     Parameters
#     ----------
#     sim_info : SimInfo
#         The simulation info object
#     """
#     metadata = {}
#     path_generator = PathGenerator(sim_info, arrays)

#     for src, mic in self.mic_src_combos():
#         self.path_info[f"{src.name}->{mic.name}"] = {}

#         if src.num == 0 or mic.num == 0:
#             reverb = "none"
#         else:
#             reverb = self.path_type[src.name][mic.name]

#         self.path_info[f"{src.name}->{mic.name}"]["type"] = reverb
#         print(f"{src.name}->{mic.name} has propagation type: {reverb}")
#         if reverb != "modified":
#             self.paths[src.name][mic.name], path_info = self.path_generator.create_path(src, mic, reverb, sim_info, True, True)
#             for key, val in path_info.items():
#                 self.path_info[f"{src.name}->{mic.name}"][key] = val

class PathGenerator:
    def __init__(self, sim_info, arrays):
        """A class in charge of generating the impulse response between sources and microphones

        Parameters
        ----------
        sim_info : SimInfo
            The simulation info object
        arrays : ArrayCollection
            The array collection object
        
        Notes
        -----
        Currently the PathGenerator is instantiated as a member of the ArrayCollections class,
        while also requiring an ArrayCollection object as an argument. This is not a natural
        dependency and should be changed in the future.
        """
        self.sim_info = sim_info
        self.arrays = arrays

        #self.filter_below = 50
        #self.filt_ir = spsig.firwin(2*int(0.05 * sim_info.samplerate) + 1, cutoff=self.filter_below, pass_zero="highpass", fs=sim_info.samplerate)

        #self.truncate_at = sim_info.max_room_ir_length

        self.calc_rir_parameters(self.sim_info)

    def calc_rir_parameters(self, sim_info):
        """Calculates the parameters used to generate the impulse response. 
        
        These parameters should be constant over the course of a simulation, and are calculated
        once at the beginning of the simulation. 

        Parameters
        ----------
        sim_info : SimInfo
            The simulation info object
        
        """
        pra.constants.set("c", sim_info.c)

        room_center = np.array(sim_info.room_center)
        room_size = np.array(sim_info.room_size)

        rt60 = sim_info.rt60
        if rt60 > 0:
            self.e_absorbtion, self.max_order = pra.inverse_sabine(rt60, room_size)
            self.max_order += 8
        else:
            self.e_absorbtion = 0.9
            self.max_order = 0

        shortest_distance = np.inf
        for src, mic in self.arrays.mic_src_combos():
            shortest_distance = np.min((shortest_distance, np.min(distfuncs.cdist(src.pos, mic.pos))))
       # shortest_distance = np.min(distfuncs.cdist(pos_from, pos_to))
        #min_dly = int(np.ceil(shortest_distance * samplerate / c))
        self.min_dly = 0
        frac_dly_len = 2*(self.min_dly + sim_info.extra_delay) + 1
        pra.constants.set("frac_delay_length",frac_dly_len)
        #if verbose:
        if frac_dly_len < 20:
            print("WARNING: fractional delay length: ",frac_dly_len)



    def create_path (self, src, mic, reverb, sim_info, return_path_info=False, verbose=False):
        """Generate the impulse response between a source and a microphone array
        
        Parameters
        ----------
        src : Array 
            The source array
        mic : Array
            The microphone array
        reverb : str
            The type of propagation between the source and the microphone array
            Possible values are "none", "direct", "random", "ism"
        sim_info : SimInfo
            The simulation info object
        return_path_info : bool, optional
            If True, the method will return a tuple of the path and a dict containing metadata
            about the path. The default is False.
        verbose : bool, optional
            If True, the method will print information about the path generation. The default is False.
        """
        path_info = {}
        if reverb == "none": 
            path = np.zeros((src.num, mic.num, 1))
        elif reverb == "direct":
            assert src.num == mic.num
            path = np.eye(src.num, mic.num)[...,None]
        elif reverb == "random":
            path = np.random.normal(size=(src.num, mic.num, sim_info.max_room_ir_length))
        elif reverb == "ism":
            if sim_info.spatial_dims == 3:
                path = ir_room_image_source_3d(
                        src.pos, 
                        mic.pos, 
                        sim_info.room_size, 
                        sim_info.room_center, 
                        sim_info.max_room_ir_length, 
                        sim_info.samplerate, 
                        self.e_absorbtion, 
                        self.max_order, 
                        self.min_dly,
                        mic.directivity_type,
                        mic.directivity_dir,
                        randomized_ism = sim_info.randomized_ism,
                        calculate_metadata=return_path_info,
                        verbose = verbose)
                if return_path_info:
                    path, path_info["ism_info"] = path
            else:
                raise ValueError
        #elif reverb == "modified":
        #    pass
        else:
            raise ValueError
        
        # ADD POST PROCESSING HERE
        if sim_info.highpass_cutoff > 0:
            path = filter_rirs(path, sim_info.samplerate, sim_info.highpass_cutoff)
        # if self.filter_below > 0:
        #     tot_len = path.shape[2] + self.filt_ir.shape[-1] - 1
        #     path_new = np.zeros((path.shape[0], path.shape[1], tot_len))
        #     for i in range(path.shape[0]):
        #         for j in range(path.shape[1]):
        #             path_new[i,j,:] = spsig.convolve(path[i,j,:], self.filt_ir, mode="full")
        #     path = path_new[:,:,:path.shape[-1]]

        if return_path_info:
            return path, path_info
        return path


def filter_rirs(rir, sr, cutoff):
    sos = spsig.butter(2, cutoff, 'highpass', fs=sr, output='sos')
    pad = sr
    rir_padded = np.concatenate((np.zeros((*rir.shape[:-1], pad)), rir, np.zeros((*rir.shape[:-1], pad))), axis=-1)
    filtered_rir = spsig.sosfiltfilt(sos, rir_padded, axis=-1)
    filtered_rir = filtered_rir[...,pad:-pad]
    return filtered_rir

def _cardoid_to_pyroomacoustics(dir_dir):
    assert dir_dir.ndim == 2
    assert dir_dir.shape[1] == 3
    
    radius, angles = gp.cart2spherical(dir_dir)
    dir_obj = pradir.CardioidFamily(
        orientation=pradir.DirectionVector(azimuth=angles[0,0], colatitude=angles[0,1], degrees=False),
        pattern_enum=pradir.DirectivityPattern.CARDIOID,
    )
    return dir_obj

def _directionality_arg_to_pyroomacoustics(dir_type, dir_dir):
    if dir_type is None:
        assert dir_dir is None
        return None

    if len(dir_type) > 1:
        if all([dt == "omni" for dt in dir_type]):
            return None
        elif all([dt == "cardioid" for dt in dir_type]):
            return [_cardoid_to_pyroomacoustics(dir_dir[i:i+1,:]) for i in range(len(dir_type))]
        raise NotImplementedError

    if dir_type[0] == "cardioid":
        dir_obj = [_cardoid_to_pyroomacoustics(dir_dir)]

    elif dir_type[0] == "omni":
        dir_obj = None
    else:
        raise NotImplementedError
    return dir_obj


def ir_room_image_source_3d(
    pos_src,
    pos_mic,
    room_size,
    room_center,
    ir_len,
    samplerate,
    e_absorbtion,
    max_order,
    min_dly,
    dir_type_mic = None,
    dir_dir_mic = None,
    randomized_ism = False,
    calculate_metadata=False,
    verbose=False,
    #extra_delay = 0, # this is multiplied by two since frac_dly must be even
    ):
    """Generates a room impulse response using the image-source method from the pyroomacoustics library.

    Parameters
    ----------
    pos_from : ndarray of shape (num_sources, 3)
        The positions of the sources
    pos_to : ndarray of shape (num_mics, 3)
        The positions of the microphones
    room_size : list of length 3 or ndarray of shape (3,)
        The size of the room in meters
    room_center : list of length 3 or ndarray of shape (3,)
        Coordinate of the center of the room
    ir_len : int
        The length of the impulse responses in samples. 
        The impulse responses will be truncated to this length.
    samplerate : int
        The sampling rate of the impulse responses
    e_absorbtion : float
        The energy absorption coefficient of the room
    max_order : int
        The maximum order of the image sources
    min_dly : int
        The smallest propagation delay between any of the sources and microphones. If this is
        known and the value used as argument, the generated RIRs will have the correct propagation delay. 
        Otherwise they will have an extra delay.
    randomized_ism : bool, optional
        If True, it will use the randomized image-source method. The default is True.
    calculate_metadata : bool, optional
        If True, the method will also return adict containing metadata
        about the impulse responses. The default is False.
    verbose : bool, optional
        If True, the method will print information about the path generation. The default is False.

    Returns
    -------
    ir : ndarray of shape (num_sources, num_mics, ir_len)
        The generated impulse responses
    metadata : dict
        A dictionary containing metadata about the impulse responses. This is only returned if
        calculate_metadata is True.
    """

    num_from = pos_src.shape[0]
    num_to = pos_mic.shape[0]
    ir = np.zeros((num_from, num_to, ir_len))

    room_center = np.array(room_center)
    room_size = np.array(room_size)
    pos_offset = room_size / 2 - room_center

    max_trunc_error = -np.inf
    max_trunc_value = -np.inf
    max_num_ir_at_once = 500
    num_computed = 0
    while num_computed < num_to:
        room = pra.ShoeBox(
            room_size,
            materials=pra.Material(e_absorbtion),
            fs=samplerate,
            max_order=max_order,
            use_rand_ism = randomized_ism, 
            max_rand_disp = 0.05
        )

        for src_idx in range(num_from):
            room.add_source((pos_src[src_idx, :] + pos_offset).T)

        block_size = np.min((max_num_ir_at_once, num_to - num_computed))
        
        if any([dir_type != "omni" for dir_type in dir_type_mic]):
            raise NotImplementedError("Directional microphones not implemented yet")
            mics = pra.MicrophoneArray(
                (pos_mic[num_computed : num_computed + block_size, :] + pos_offset[None, :]).T,
                samplerate,
                directivity=dir_arg[num_computed : num_computed + block_size],
            )
        else:
            mics = pra.MicrophoneArray(
            (pos_mic[num_computed : num_computed + block_size, :] + pos_offset[None, :]).T,
            samplerate
        )
        room.add_microphone_array(mics)

        if verbose:
            print(
                "Computing RIR {} - {} of {}".format(
                    num_computed * num_from + 1,
                    (num_computed + block_size) * num_from,
                    num_to * num_from,
                )
            )
        room.compute_rir()
        for to_idx, receiver in enumerate(room.rir):
            for from_idx, single_rir in enumerate(receiver):
                ir_len_to_use = np.min((len(single_rir), ir_len)) - min_dly
                ir[from_idx, num_computed + to_idx, :ir_len_to_use] = np.array(single_rir)[
                    min_dly:ir_len_to_use+min_dly
                ]
        num_computed += block_size

        if calculate_metadata:
            truncError, truncValue = calc_truncation_info(room.rir, ir_len)
            max_trunc_error = np.max((max_trunc_error, truncError))
            max_trunc_value = np.max((max_trunc_value, truncValue))

    if calculate_metadata:
        metadata = {}
        metadata["Max Normalized Truncation Error (dB)"] = max_trunc_error
        metadata["Max Normalized Truncated Value (dB)"] = max_trunc_value
        if max_order > 0:
            try:
                metadata["Measured RT60 (min)"] = np.min(room.measure_rt60())
                metadata["Measured RT60 (max)"] = np.max(room.measure_rt60())
            except ValueError:
                metadata["Measured RT60 (min)"] = "failed"
                metadata["Measured RT60 (max)"] = "failed"
        else:
            metadata["Measured RT60 (min)"] = 0
            metadata["Measured RT60 (max)"] = 0
        metadata["Max ISM order"] = max_order
        metadata["Energy Absorption"] = e_absorbtion
        return ir, metadata
    return ir




def ir_room_image_source_3d_orig(
    pos_from,
    pos_to,
    room_size,
    room_center,
    ir_len,
    rt60,
    samplerate,
    c,
    randomized_ism = True,
    calculate_metadata=False,
    verbose=False,
    extra_delay = 0, # this is multiplied by two since frac_dly must be even
):

    num_from = pos_from.shape[0]
    num_to = pos_to.shape[0]
    ir = np.zeros((num_from, num_to, ir_len))
    room_center = np.array(room_center)
    room_size = np.array(room_size)

    pos_offset = room_size / 2 - room_center

    if rt60 > 0:
        e_absorbtion, max_order = pra.inverse_sabine(rt60, room_size)
        max_order += 8
    else:
        e_absorbtion = 0.9
        max_order = 0
    # print("Energy Absorption: ", eAbsorption)
    # print("Max Order: ", maxOrder)

    pra.constants.set("c", c)

    shortest_distance = np.min(distfuncs.cdist(pos_from, pos_to))
    #min_dly = int(np.ceil(shortest_distance * samplerate / c))
    min_dly = 0
    frac_dly_len = 2*(min_dly + extra_delay) + 1
    pra.constants.set("frac_delay_length",frac_dly_len)
    if verbose:
        if frac_dly_len < 20:
            print("WARNING: fractional delay length: ",frac_dly_len)

    max_trunc_error = -np.inf
    max_trunc_value = -np.inf
    max_num_ir_at_once = 500
    num_computed = 0
    while num_computed < num_to:
        room = pra.ShoeBox(
            room_size,
            materials=pra.Material(e_absorbtion),
            fs=samplerate,
            max_order=max_order,
            use_rand_ism = randomized_ism, 
            max_rand_disp = 0.05
        )
        # room = pra.ShoeBox(roomSim, materials=pra.Material(e_absorption), fs=sampleRate, max_order=max_order)

        for src_idx in range(num_from):
            room.add_source((pos_from[src_idx, :] + pos_offset).T)

        block_size = np.min((max_num_ir_at_once, num_to - num_computed))
        mics = pra.MicrophoneArray(
            (pos_to[num_computed : num_computed + block_size, :] + pos_offset[None, :]).T,
            room.fs,
        )
        room.add_microphone_array(mics)

        if verbose:
            print(
                "Computing RIR {} - {} of {}".format(
                    num_computed * num_from + 1,
                    (num_computed + block_size) * num_from,
                    num_to * num_from,
                )
            )
        room.compute_rir()
        for to_idx, receiver in enumerate(room.rir):
            for from_idx, single_rir in enumerate(receiver):
                ir_len_to_use = np.min((len(single_rir), ir_len)) - min_dly
                ir[from_idx, num_computed + to_idx, :ir_len_to_use] = np.array(single_rir)[
                    min_dly:ir_len_to_use+min_dly
                ]
        num_computed += block_size

        if calculate_metadata:
            truncError, truncValue = calc_truncation_info(room.rir, ir_len)
            max_trunc_error = np.max((max_trunc_error, truncError))
            max_trunc_value = np.max((max_trunc_value, truncValue))

    if calculate_metadata:
        metadata = {}
        metadata["Max Normalized Truncation Error (dB)"] = max_trunc_error
        metadata["Max Normalized Truncated Value (dB)"] = max_trunc_value
        if rt60 > 0:
            try:
                metadata["Measured RT60 (min)"] = np.min(room.measure_rt60())
                metadata["Measured RT60 (max)"] = np.max(room.measure_rt60())
            except ValueError:
                metadata["Measured RT60 (min)"] = "failed"
                metadata["Measured RT60 (max)"] = "failed"
        else:
            metadata["Measured RT60 (min)"] = 0
            metadata["Measured RT60 (max)"] = 0
        metadata["Max ISM order"] = max_order
        metadata["Energy Absorption"] = e_absorbtion
        return ir, metadata
    return ir

    
def calc_truncation_info(all_rir, trunc_len):
    max_trunc_error = -np.inf
    max_trunc_value = -np.inf
    for to_idx, receiver in enumerate(all_rir):
        for from_idx, single_rir in enumerate(receiver):
            ir_len = np.min((len(single_rir), trunc_len))
            trunc_error = calc_truncation_error(single_rir, ir_len)
            trunc_value = calc_truncation_value(single_rir, ir_len)
            max_trunc_error = np.max((max_trunc_error, trunc_error))
            max_trunc_value = np.max((max_trunc_value, trunc_value))
    return max_trunc_error, max_trunc_value


def calc_truncation_error(rir, trunc_len):
    tot_power = np.sum(rir ** 2)
    trunc_power = np.sum(rir[trunc_len:] ** 2)
    trunc_error = trunc_power / tot_power
    if trunc_error == 0:
        return -np.inf
    return 10 * np.log10(trunc_error)


def calc_truncation_value(rir, trunc_len):
    if len(rir) <= trunc_len:
        return -np.inf
    max_trunc_value = np.max(np.abs(rir[trunc_len:]))
    max_value = np.max(np.abs(rir))
    return 20 * np.log10(max_trunc_value / max_value)


def show_rt60(multi_channel_ir):
    for i in range(multi_channel_ir.shape[0]):
        for j in range(multi_channel_ir.shape[1]):
            single_ir = multi_channel_ir[i, j, :]
            rt = pra.experimental.rt60.measure_rt60(single_ir)
            print("rt60 is: ", rt)
